from abc import ABCMeta, abstractmethod
from pathlib import Path
import sys
from typing import (
    Any, Dict, IO, List, Optional, Sequence as SequenceType, Set, Tuple, Type, Union
)

from xphyle import STDOUT, xopen, open_
from xphyle.types import ModeArg

from atropos.commands.trim.filters import Filter, NoFilter
from atropos.io.sequence import Sequence
from atropos.io.formatters import Formatter, create_seq_formatter
from atropos.utils.paths import splitext_compressed
from atropos.utils.collections import Summarizable


FileDesc = Union[Path, Tuple[Path, ModeArg]]


class Writers:
    """
    Manages writing to one or more outputs.
    """

    def __init__(
        self,
        force_create: SequenceType[Path] = (),
        compression_format: Optional[str] = None
    ):
        """
        Args:
            force_create: Whether empty output files should be created.
        """
        self.writers = {}
        self.force_create = force_create
        self.suffix = None
        self.compression_format = compression_format

    def get_writer(self, file_desc: FileDesc, compressed: bool = False) -> IO:
        """
        Create the writer for a file descriptor if it does not already exist.

        Args:
            file_desc: File descriptor. If `compressed==True`, this is a tuple
                (path, mode), otherwise it's only a path.
            compressed: Whether data has already been compressed.

        Returns:
            The writer.
        """
        if compressed:
            path, mode = file_desc
            compression = False
        else:
            path = file_desc
            mode = "w"
            compression = self.compression_format

        if path not in self.writers:
            if self.suffix:
                real_path = add_suffix_to_path(path, self.suffix)
            else:
                real_path = path

            # TODO: test whether O_NONBLOCK allows non-blocking write to NFS
            self.writers[path] = xopen(real_path, mode, compression=compression)

        return self.writers[path]

    def write_result(self, result, compressed: bool = False) -> None:
        """
        Writes results to output.

        Args:
            result: Dict with keys being file descriptors and values being data
                (either bytes or strings). Strings are expected to already have
                appropriate line-endings.
            compressed: Whether data has already been compressed.
        """
        for file_desc, data in result.items():
            self.write(file_desc, data, compressed)

    def write(self, file_desc: FileDesc, data, compressed: bool = False):
        """
        Writes data to output. If the specified path has not been seen before,
        the output is opened.

        Args:
            file_desc: File descriptor. If `compressed==True`, this is a tuple
                (path, mode), otherwise it's only a path.
            data: The data to write.
            compressed: Whether data has already been compressed.
        """
        self.get_writer(file_desc, compressed).write(data)

    def close(self):
        """
        Closes all outputs.
        """
        for path in self.force_create:
            if path not in self.writers and path != STDOUT:
                with open_(path, "w"):
                    pass

        for writer in self.writers.values():
            if writer not in (sys.stdout, sys.stderr):
                writer.close()


class DetailFormatter(metaclass=ABCMeta):
    """
    Base class for formatters that write to a delimited file.

    Args:
        path: The output file path.
        delim: The field delimiter.
    """

    def __init__(self, path, delim=" "):
        self.path = path
        self.delim = delim

    @abstractmethod
    def format(self, result: dict, read: Sequence):
        """
        Format a read and add it to `result`.
        """

    def _format(self, result: dict, fields: SequenceType[Any]):
        if self.path not in result:
            result[self.path] = []

        result[self.path].append(
            "".join((self.delim.join(str(f) for f in fields), "\n"))
        )


class RestFormatter(DetailFormatter):
    """
    Rest file formatter.
    """

    def format(self, result: dict, read: Sequence):
        if read.match:
            rest = read.match.rest()

            if len(rest) > 0:
                self._format(result, (rest, read.name))


class InfoFormatter(DetailFormatter):
    """
    Info file formatter.
    """

    def __init__(self, path: Path):
        super().__init__(path, delim="\t")

    def format(self, result: dict, read: Sequence):
        if read.match:
            for match_info in read.match_info:
                self._format(result, match_info[0:11])
        else:
            seq = read.sequence
            qualities = read.qualities if read.qualities is not None else ""
            self._format(result, (read.name, -1, seq, qualities))


class WildcardFormatter(DetailFormatter):
    """
    Wildcard file formatter.
    """

    def format(self, result: dict, read: Sequence):
        if read.match:
            self._format(result, (read.match.wildcards(), read.name))


class Formatters(Summarizable):
    """
    Manages multiple formatters.
    """

    def __init__(self, output: Path, seq_formatter_args: dict):
        """
        Args:
            output: The output file name template.
            seq_formatter_args: Additional arguments to pass to the formatter
                constructor.
        """
        self.output = output
        self.multiplexed = output is not None and "{name}" in output.name
        self.seq_formatter_args = seq_formatter_args
        self.seq_formatters: Dict[Type[Filter], Formatter] = {}
        self.mux_formatters: Dict[str, Formatter] = {}
        self.detail_formatters: List[DetailFormatter] = []
        self.discarded = 0

    def add_seq_formatter(
        self, filter_type: Type[Filter], file1: Path, file2: Optional[Path] = None
    ):
        """
        Adds a formatter.

        Args:
            filter_type: The type of filter that triggers writing with the
                formatter.
            file1:
            file2: The output file(s).
        """
        self.seq_formatters[filter_type] = create_seq_formatter(
            file1, file2, **self.seq_formatter_args
        )

    def add_detail_formatter(self, formatter: DetailFormatter):
        """
        Add a formatter for one of the delimited detail files (rest, info, wildcard).
        """
        self.detail_formatters.append(formatter)

    def get_mux_formatter(self, name: str) -> Formatter:
        """
        Returns the formatter associated with the given name (barcode) when running
        in multiplexed mode.
        """
        if not self.multiplexed:
            raise ValueError(
                "Cannot call 'get_mux_formatter' for non-multiplexed data"
            )

        if name not in self.mux_formatters:
            path = Path(str(self.output).format(name=name))
            self.mux_formatters[name] = create_seq_formatter(
                path, **self.seq_formatter_args
            )

        return self.mux_formatters[name]

    def get_seq_formatters(self) -> Set[Formatter]:
        """
        Returns a set containing all formatters that have handled at least one record.
        """
        return set(f for f in self.seq_formatters.values() if f.written > 0) | set(
            f for f in self.mux_formatters.values() if f.written > 0
        )

    def format(
        self,
        result: dict,
        dest: Type[Filter],
        read1: Sequence,
        read2: Optional[Sequence] = None
    ):
        """
        Formats read(s) and add to a result dict. Also writes info records to any
        registered info formatters.

        Args:
            result: The result dict.
            dest: The destination (filter type).
            read1: The first read.
            read2: The second read.
        """
        if self.multiplexed and (dest == NoFilter) and read1.match:
            name = read1.match.adapter.name
            formatter = self.get_mux_formatter(name)
            formatter.format(result, read1, read2)
        elif dest in self.seq_formatters:
            self.seq_formatters[dest].format(result, read1, read2)
        else:
            self.discarded += 1

        for fmtr in self.detail_formatters:
            fmtr.format(result, read1)
            if read2:
                fmtr.format(result, read2)

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        seq_formatters = self.get_seq_formatters()
        return dict(
            records_written=sum(f.written for f in seq_formatters),
            bp_written=[
                sum(f.read1_bp for f in seq_formatters),
                sum(f.read2_bp for f in seq_formatters),
            ],
        )


def add_suffix_to_path(path: Path, suffix: str) -> str:
    """
    Add the suffix (str or int) after the file name but
    before the extension.
    """
    name, ext1, ext2 = splitext_compressed(path)
    return f"{name}{suffix}{ext1}{ext2 or ''}"



from atropos.commands.console import (
    BaseCommandConsole,
    add_common_options,
    validate_common_options,
    parse_metrics_args,
    configure_threads,
)
from atropos.commands.qc import QcCommand
from atropos.commands.legacy_reports import LegacyReportGenerator
from atropos.utils import classproperty
from atropos.utils.argparse import (
    AtroposArgumentParser,
    Namespace,
    positive,
    int_or_str,
    writable_file,
)


class QcCommandConsole(QcCommand, LegacyReportGenerator, BaseCommandConsole):
    @classproperty
    def description(cls) -> str:
        return """
        atropos qc -se input.fastq
        atropos qc -pe1 in1.fastq -pe2 in2.fastq
        """

    @classproperty
    def usage(cls) -> str:
        return """
        Compute read-level statistics. The output is identical to running the 'trim'
        command with '--metrics pre'. Use this command to get an idea of the quality of
        your raw data.
        """

    @classmethod
    def _add_arguments(cls, parser: AtroposArgumentParser) -> None:
        add_common_options(parser)
        cls._add_qc_options(parser)

    @classmethod
    def _validate_options(
        cls, options: Namespace, parser: AtroposArgumentParser
    ) -> None:
        validate_common_options(options, parser)
        cls._validate_qc_options(options, parser)

    @classmethod
    def _add_qc_options(cls, parser: AtroposArgumentParser):
        parser.set_defaults(action="qc", batch_size=None)
        group = parser.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writable_file,
            default="-",
            metavar="FILE",
            help="Write metrics to file rather than stdout.",
        )
        group = parser.add_group(
            "Report", title="Report content and formatting options"
        )
        group.add_argument(
            "--report-formats",
            nargs="*",
            choices=cls.list_report_formats(),
            default=None,
            metavar="FORMAT",
            help="Report type(s) to generate. If multiple, '--output' "
            "is treated as a prefix and the appropriate extensions are "
            "appended. If unspecified, the format is guessed from the "
            "output file extension.",
        )
        group.add_argument(
            "--metrics",
            type=parse_metrics_args,
            default=None,
            help="Additional arguments for read statistic collection. E.g. "
            "'pre:tiles' means to also collect tile-level statistics "
            "(Illumina data only), and 'pre:tiles=<regexp>' means to use "
            "the specified regular expression to extract key portions of "
            "read names to collect the tile statistics.",
        )
        group = parser.add_group("Parallel", title="Parallel (multi-core) options")
        group.add_argument(
            "-T",
            "--threads",
            type=positive(int, True),
            default=None,
            metavar="THREADS",
            help="Number of threads to use for read trimming. Set to 0 to use "
            "max available threads. (Do not use multithreading)",
        )
        group.add_argument(
            "--process-timeout",
            type=positive(int, True),
            default=60,
            metavar="SECONDS",
            help="Number of seconds process should wait before escalating "
            "messages to ERROR level. (60)",
        )
        group.add_argument(
            "--read-queue-size",
            type=int_or_str,
            default=None,
            metavar="SIZE",
            help="Size of queue for batches of reads to be processed. "
            "(THREADS * 100)",
        )

    @staticmethod
    def _validate_qc_options(options: Namespace, parser: AtroposArgumentParser):
        options.report_file = options.output

        if options.threads is not None:
            threads = configure_threads(options, parser)
            if options.read_queue_size is None:
                options.read_queue_size = threads * 100
            elif 0 < options.read_queue_size < threads:
                parser.error("Read queue size must be >= than 'threads'")

        if options.batch_size is None:
            options.batch_size = 1000



from abc import ABCMeta, abstractmethod
from argparse import (
    Action, ArgumentParser, ArgumentTypeError, HelpFormatter, Namespace,
)
from numbers import Number
import operator
from pathlib import Path
import re
import textwrap
from typing import (
    Any,
    Callable,
    Sequence,
    List,
    Tuple,
    Type,
    TypeVar,
    Generic,
    Union,
    Optional,
    Text,
)
from urllib.parse import urlparse

from xphyle.paths import (
    STDOUT,
    STDERR,
    check_path,
    as_path,
    resolve_path,
    check_readable_file,
    check_writable_file,
)
from xphyle.types import PathLike, PathType, Permission

from atropos.utils import Magnitude


T = TypeVar("T")


class EnumAction(Action, metaclass=ABCMeta):
    """
    This action performs conversion from string to enum. This is done in an action
    rather than by using a type conversion to enable enum choices to be printed as
    string values.
    """

    def __init__(
        self, const: Union[Type[T], Tuple[Type[T], Sequence[T]]], **kwargs
    ) -> None:
        if isinstance(const, tuple):
            self._enum_class = const[0]
            kwargs["choices"] = self._get_choices(const[1])
        else:
            self._enum_class = const
            kwargs["choices"] = self._get_choices(const)

        if "metavar" not in kwargs or kwargs["metavar"] is None:
            kwargs["metavar"] = self._enum_class.__name__.upper()

        super().__init__(**kwargs)

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[Text, Sequence[Any], None],
        option_string: Optional[Text] = None
    ) -> None:
        if values is None:
            enum_value = None
        elif isinstance(values, Text):
            enum_value = self._to_enum_value(values)
        else:
            enum_value = tuple(self._to_enum_value(value) for value in values)

        setattr(namespace, self.dest, enum_value)

    @abstractmethod
    def _get_choices(self, choices: Sequence[T]) -> Tuple[str]:
        pass

    @abstractmethod
    def _to_enum_value(self, value):
        pass


class EnumNameAction(EnumAction):
    def _get_choices(self, choices: Sequence[T]) -> Tuple[str]:
        return tuple(str(c.name).lower().replace("_", "-") for c in choices)

    def _to_enum_value(self, value):
        return self._enum_class[value.upper().replace("-", "_")]


class EnumValueAction(EnumAction):
    def _get_choices(self, choices: Sequence[T]) -> Tuple[str]:
        return tuple(str(c.value) for c in choices)

    def _to_enum_value(self, value):
        return self._enum_class(value)


class AtroposArgumentParser(ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._groups = {}

    def add_group(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        mutex: bool = False,
        required: bool = False,
    ):
        """
        Add a group to the parser. The group will be stored under `name` and can
        later be retrieved via `get_group`.
        """
        if name in self._groups:
            raise ValueError(f"Group already exists: {name}")

        if mutex:
            group = self.add_mutually_exclusive_group(required=required)
        else:
            group = self.add_argument_group(title or name, description)
        self._groups[name] = group
        return group

    def get_group(self, name: str):
        """
        If a group has already been created with `name`, return the group, otherwise
        create a new group with that name.
        """
        if name in self._groups:
            return self._groups[name]
        else:
            return self.add_group(name)


class ParagraphHelpFormatter(HelpFormatter):
    """
    HelpFormatter that wraps text in paragraphs.
    """

    def _fill_text(self, text, width, indent):
        text = re.sub("[ \t]{2,}", " ", text)
        paragraphs = [
            textwrap.fill(p, width, initial_indent=indent, subsequent_indent=indent)
            for p in re.split("\n\n", text)
        ]
        return "\n\n".join(paragraphs)


class CompositeType:
    """A composite of multiple data types.
    """

    def __init__(self, *types):
        self.types = types

    def __call__(self, string):
        result = string
        for datatype in self.types:
            result = datatype(result)
        return result


class ComparisonValidator(Generic[T]):
    """
    Validator that compares an argument against an expected value.
    """

    def __init__(self, rhs: T, oper: Callable[[T, T], bool], expected: bool = True):
        self.rhs = rhs
        self.oper = oper
        self.expected = expected

    def __call__(self, lhs: T):
        if self.oper(lhs, self.rhs) != self.expected:
            raise ArgumentTypeError(
                f"{self.oper}({lhs}, {self.rhs}) != {self.expected}"
            )

        return lhs


class Delimited(Generic[T]):
    """
    Splits a string argument using a delimiter.
    """

    def __init__(
        self,
        delim: str = ",",
        data_type: Optional[Callable[[str], T]] = None,
        choices: Optional[Sequence[Union[str, T]]] = None,
        min_len: Optional[int] = None,
        max_len: Optional[int] = None,
    ):
        self.delim = delim
        self.data_type = data_type
        self.choices = choices
        self.min_len = min_len
        self.max_len = max_len

    def __call__(self, arg: Union[str, List[T]]) -> List[T]:
        if isinstance(arg, str):
            vals = arg.split(self.delim) if self.delim else [arg]
        else:
            vals = arg
        if vals[0] == "*" and self.choices is not None:
            vals = self.choices
        if self.data_type:
            vals = [self.data_type(v) for v in vals]

        if self.min_len and len(vals) < self.min_len:
            raise ArgumentTypeError(f"There must be at least {self.min_len} values")

        if self.max_len and len(vals) > self.max_len:
            raise ArgumentTypeError(f"There can be at most {self.max_len} values")

        return vals


class EnumCharList:
    """
    Parses a string into a list of characters and then converts each into an enum value.
    """

    def __init__(self, enum_class: Callable[[str], T]):
        self.enum_class = enum_class

    def __call__(self, string: str) -> Sequence[T]:
        try:
            return [self.enum_class(c) for c in string]
        except KeyError:
            raise ArgumentTypeError(
                f"One or more characters in {string} not a {self.enum_class}"
            )


class AccessiblePath:
    """
    Tests that a path is accessible.
    """

    def __init__(self, path_type: PathType, mode: Permission):
        self.path_type = path_type
        self.mode = mode

    def __call__(self, path: Union[str, PathLike]) -> Path:
        path = as_path(path)

        if self.path_type == PathType.FILE and path in (STDOUT, STDERR):
            return path

        try:
            return Path(check_path(path, self.path_type, self.mode))
        except IOError:
            raise ArgumentTypeError(
                f"Path {path} is not an accessible {self.path_type} with mode "
                f"{self.mode}"
            )


def existing_path(path: Union[str, PathLike]) -> Path:
    """
    Tests that a path exists.
    """
    path = as_path(path)

    if path == STDOUT:
        return path

    try:
        return Path(resolve_path(path))
    except IOError:
        raise ArgumentTypeError(f"Path {path} does not exist")


def readable_file(path):
    try:
        return check_readable_file(resolve_path(path))
    except IOError:
        raise ArgumentTypeError(
            f"Path {path} does not exist, is not a file, or is not readable"
        )


def writable_file(path):
    try:
        return check_writable_file(as_path(path, "w"))
    except IOError:
        raise ArgumentTypeError(f"Path {path} is not writable")


def readwritable_file(path):
    """Tests that a file is both readable and writeable."""
    path = as_path(path)

    if path.exists():
        path = readable_file(path)

    return writable_file(path)


def readable_url(url):
    """
    Validator for a URL that must be readable.

    Args:
        url: The URL to validate.
    """
    parsed = urlparse(url)
    scheme = parsed.scheme or "file"
    if scheme == "file":
        filename = readable_file(parsed.path)
        return "file:" + str(filename)
    else:
        return url


str_list = Delimited[str](data_type=str)
"""Comma-delimited list of strings."""

INT_OR_STR_RE = re.compile(r"([\d.]+)([KkMmGg]?)")


def int_or_str(arg):
    """
    Similar to int(), but accepts K, M, and G abbreviations.
    """
    if arg is None or isinstance(arg, int):
        return arg
    elif isinstance(arg, str):
        match = INT_OR_STR_RE.match(arg.upper())
        num, mult = match.groups()
        if mult:
            return int(float(num) * Magnitude[mult].divisor)
        else:
            return int(num)
    else:
        raise ArgumentTypeError(f"Unsupported type {arg}")


def positive(type_=int, inclusive=False):
    """
    Tests that a number is greater than (or equal to, if ``inclusive=True``) zero.
    """
    oper = operator.ge if inclusive else operator.gt
    return CompositeType(type_, ComparisonValidator(0, oper))


N = TypeVar("N", bound=Number)


def between(min_val: N = None, max_val: N = None, type_: Type[N] = int):
    """
    Returns a CompositeType that validates `min_val <= x <= max_val`.
    """
    return CompositeType(
        type_,
        ComparisonValidator(min_val, operator.ge),
        ComparisonValidator(max_val, operator.le),
    )


probability = between(0, 1, float)
"""A float between 0-1 (inclusive)."""



from subprocess import CalledProcessError
import urllib.request

import subby


def assert_files_equal(path1, path2):
    try:
        subby.sub("diff -u {0} {1}".format(path1, path2))
    except CalledProcessError as e:
        raise AssertionError(
            f"Files not equal: {path1} != {path2}\nDiff: <{e.output}>"
        )


def no_internet(url="https://github.com"):
    """
    Tests whether there's no internet connection available.
    """
    try:
        urllib.request.urlopen(url).info()
        return False
    except:
        return True



from atropos.commands.trim.filters import (
    NContentFilter,
    DISCARD,
    KEEP,
    SingleWrapper,
    PairedWrapper,
)
from atropos.io.sequence import Sequence


def test_ncontentfilter():
    # third parameter is True if read should be discarded
    params = [
        ("AAA", 0, KEEP),
        ("AAA", 1, KEEP),
        ("AAACCTTGGN", 1, KEEP),
        ("AAACNNNCTTGGN", 0.5, KEEP),
        ("NNNNNN", 1, DISCARD),
        ("ANAAAA", 1 / 6, KEEP),
        ("ANAAAA", 0, DISCARD),
    ]
    for seq, count, expected in params:
        _filter = NContentFilter(count=count)
        _seq = Sequence("read1", seq, qualities="#" * len(seq))
        assert _filter(_seq) == expected


def test_ncontentfilter_paired():
    params = [
        ("AAA", "AAA", 0, KEEP),
        ("AAAN", "AAA", 0, DISCARD),
        ("AAA", "AANA", 0, DISCARD),
        ("ANAA", "AANA", 1, KEEP),
    ]
    for seq1, seq2, count, expected in params:
        _filter = NContentFilter(count=count)
        filter_legacy = SingleWrapper(_filter)
        filter_both = PairedWrapper(_filter)
        read1 = Sequence("read1", seq1, qualities="#" * len(seq1))
        read2 = Sequence("read1", seq2, qualities="#" * len(seq2))
        assert filter_legacy(read1, read2) == _filter(read1)
        # discard entire pair if one of the reads fulfills criteria
        assert filter_both(read1, read2) == expected



import shutil

import pytest
from xphyle import open_

from atropos.console import execute_cli
from atropos.commands.trim import AlignerType
from atropos.commands.trim.console import TrimCommandConsole
from atropos.utils import ReturnCode


BACK_ALIGNERS = tuple(
    a.name.lower() for a in (AlignerType.ADAPTER, AlignerType.INSERT)
)


def test_paired_separate(run_trimmer):
    run_trimmer(
        "-a TTAGACATAT",
        "paired-separate.1.fastq",
        "paired.1.fastq",
    )
    run_trimmer(
        "-a CAGTGGAGTA",
        "paired-separate.2.fastq",
        "paired.2.fastq",
    )


def test_paired_end_legacy(run_trimmer):
    # The -m 14 filters out one read, which should then also be filtered out in
    # the second output file.
    run_trimmer(
        "-a TTAGACATAT -m 14",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired.m14.1.fastq",
        expected2="paired.m14.2.fastq",
    )


def test_untrimmed_paired_output(run_trimmer):
    run_trimmer(
        [
            "-a",
            "TTAGACATAT",
            "--untrimmed-output",
            "paired-untrimmed.1.fastq",
            "--untrimmed-paired-output",
            "paired-untrimmed.2.fastq",
        ],
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired-trimmed.1.fastq",
        expected2="paired-trimmed.2.fastq",
        expected_other=("paired-untrimmed.1.fastq", "paired-untrimmed.2.fastq")
    )


def test_explicit_format_with_paired(run_trimmer, input_data, tmp_path):
    # Use --format=fastq with input files whose extension is .txt
    txt1 = tmp_path / "paired.1.txt"
    txt2 = tmp_path / "paired.2.txt"
    shutil.copyfile(input_data("paired.1.fastq"), txt1)
    shutil.copyfile(input_data("paired.2.fastq"), txt2)
    run_trimmer(
        "--input-format=fastq -a TTAGACATAT -m 14",
        inpath1=txt1,
        inpath2=txt2,
        expected1="paired.m14.1.fastq",
        expected2="paired.m14.2.fastq",
    )


def test_no_trimming_legacy(input_data):
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            "-a",
            "XXXXX",
            "-o",
            "/dev/null",
            "-p",
            "/dev/null",
            "-pe1",
            input_data("paired.1.fastq"),
            "-pe2",
            input_data("paired.2.fastq"),
        ]
    )


def test_no_trimming(input_data):
    # make sure that this doesn't divide by zero
    execute_cli(
        [
            "-a",
            "XXXXX",
            "-A",
            "XXXXX",
            "-o",
            "/dev/null",
            "-p",
            "/dev/null",
            "-pe1",
            input_data("paired.1.fastq"),
            "-pe2",
            input_data("paired.2.fastq"),
        ]
    )


def test_missing_file(input_data):
    assert ReturnCode.ERROR == execute_cli(
        ["-a", "XX", "--paired-output", "out.fastq", input_data("paired.1.fastq")]
    )


def test_first_too_short(run_trimmer, input_data, tmp_path):
    trunc1 = tmp_path / "truncated.1.fastq"
    out = tmp_path / "out.fastq"
    # Create a truncated file in which the last read is missing
    with open_(input_data("paired.1.fastq")) as f:
        lines = f.readlines()
        lines = lines[:-4]
    with open_(trunc1, "w") as f:
        f.writelines(lines)
    assert ReturnCode.ERROR == execute_cli([
        "-a", "XX", "--paired-output", out, trunc1, input_data("paired.2.fastq")
    ])


def test_second_too_short(run_trimmer, input_data, tmp_path):
    trunc2 = tmp_path / "truncated.2.fastq"
    out = tmp_path / "out.fastq"
    # Create a truncated file in which the last read is missing
    with open_(input_data("paired.2.fastq")) as f:
        lines = f.readlines()
        lines = lines[:-4]
    with open_(trunc2, "w") as f:
        f.writelines(lines)
    assert ReturnCode.ERROR == execute_cli([
        "-a", "XX", "--paired-output", out, input_data("paired.1.fastq"), trunc2
    ])


def test_unmatched_read_names(run_trimmer, input_data, tmp_path):
    swapped = tmp_path / "swapped.1.fastq"
    out1 = tmp_path / "out1.fastq"
    out2 = tmp_path / "out2.fastq"

    # Create a file in which reads 2 and are swapped
    with open_(input_data("paired.1.fastq")) as f:
        lines = f.readlines()
        lines = lines[0:4] + lines[8:12] + lines[4:8] + lines[12:]
    with open_(swapped, "w") as f:
        f.writelines(lines)
    result = TrimCommandConsole.execute([
        "-a", "XX", "-o", str(out1), "-p", str(out2), "-pe1", str(swapped), "-pe2",
        str(input_data("paired.2.fastq"))
    ])
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] != 0


def test_legacy_minlength(run_trimmer):
    """Ensure -m is not applied to second read in a pair in legacy mode"""
    run_trimmer(
        "-a XXX -m 27",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired-m27.1.fastq",
        expected2="paired-m27.2.fastq",
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_paired_end(run_trimmer, aligner):
    """single-pass paired-end with -m"""
    run_trimmer(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired_{aligner}.1.fastq",
        expected2="paired_{aligner}.2.fastq",
        aligner=aligner
    )


def test_paired_anchored_back_no_indels(run_trimmer):
    run_trimmer(
        "-a BACKADAPTER$ -A BACKADAPTER$ -N --no-indels",
        inpath1="anchored-back.fasta",
        inpath2="anchored-back.fasta",
        expected1="anchored-back.fasta",
        expected2="anchored-back.fasta",
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_paired_end_qualtrim(run_trimmer, aligner):
    """single-pass paired-end with -q and -m"""
    run_trimmer(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="pairedq.1.fastq",
        expected2="pairedq.2.fastq",
        aligner=aligner,
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_paired_end_qualtrim_swapped(run_trimmer, aligner):
    """single-pass paired-end with -q and -m, but files swapped"""
    run_trimmer(
        "-q 20 -a CAGTGGAGTA -A TTAGACATAT -m 14 --adapter-max-rmp 0.001",
        inpath1="paired.2.fastq",
        inpath2="paired.1.fastq",
        expected1="pairedq.2.fastq",
        expected2="pairedq.1.fastq",
        aligner=aligner,
    )


def test_paired_end_cut(run_trimmer):
    run_trimmer(
        "-u 3 -u -1 -U 4 -U -2",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="pairedu.1.fastq",
        expected2="pairedu.2.fastq",
    )


def test_paired_end_a_only(run_trimmer):
    run_trimmer(
        "-A CAGTGGAGTA",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired-onlyA.1.fastq",
        expected2="paired-onlyA.2.fastq",
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_paired_end_mask_adapter(run_trimmer, aligner):
    """mask adapter with N (reads maintain the same length)"""
    run_trimmer(
        "-a CAAG -A TCGA -n 3 --mask-adapter",
        inpath1="back_repeat.1.fastq",
        inpath2="back_repeat.2.fastq",
        expected1="back_repeat.1.fastq",
        expected2="back_repeat.2.fastq",
        aligner=aligner,
    )


def test_discard_untrimmed(run_trimmer):
    # issue #146
    # the first adapter is a sequence cut out from the first read
    run_trimmer(
        "-a CTCCAGCTTAGACATATC -A XXXXXXXX --discard-untrimmed",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="empty.fastq",
        expected2="empty.fastq",
    )


def test_discard_trimmed(run_trimmer):
    run_trimmer(
        "-A C -O 1 --discard-trimmed",  # applies everywhere
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="empty.fastq",
        expected2="empty.fastq",
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_interleaved(run_trimmer, aligner):
    """single-pass interleaved paired-end with -q and -m"""
    run_trimmer(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        inpath1="interleaved.fastq",
        expected1="interleaved.fastq",
        interleaved_input=True,
        interleaved_output=True,
        aligner=aligner,
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_interleaved_stdout(run_trimmer, aligner):
    """single-pass interleaved paired-end with -q and -m"""
    run_trimmer(
        "-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90",
        inpath1="interleaved.fastq",
        expected1="interleaved.fastq",
        interleaved_input=True,
        interleaved_output=True,
        aligner=aligner,
        stdout=True,
    )


def test_interleaved_no_paired_output(tmp_path):
    p1 = tmp_path / "temp-paired.1.fastq"
    p2 = tmp_path / "temp-paired.2.fastq"
    params = "-a XX --interleaved".split()
    params += ["-o", p1, "-p1", p2, "paired.1.fastq", "paired.2.fastq"]
    assert execute_cli(params) == ReturnCode.ERROR


# TODO
# def test_interleaved_input_paired_output():
#     '''single-pass interleaved paired-end with -q and -m, paired output'''
#     run_interleaved2('-q 20 -a TTAGACATAT -A CAGTGGAGTA -m 14 -M 90',
#         inpath1='interleaved.fastq', expected1='pairedq1.fastq',
#         expected2='pairedq2.fastq'
#     )
@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_pair_filter(run_trimmer, aligner):
    run_trimmer(
        "--pair-filter=both -a TTAGACATAT -A GGAGTA -m 14",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired-filterboth_{aligner}.1.fastq",
        expected2="paired-filterboth_{aligner}.2.fastq",
        aligner=aligner,
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_too_short_paired_output(run_trimmer, aligner, expected_data):
    run_trimmer(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output "
        "paired-too-short.1.fastq --too-short-paired-output paired-too-short.2.fastq",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired_{aligner}.1.fastq",
        expected2="paired_{aligner}.2.fastq",
        aligner=aligner,
        expected_other=("paired-too-short.1.fastq", "paired-too-short.2.fastq")
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_too_long_output(run_trimmer, aligner):
    run_trimmer(
        f"-a TTAGACATAT -A CAGTGGAGTA -M 14 --too-long-output "
        f"paired_{aligner}.1.fastq --too-long-paired-output paired_{aligner}.2.fastq",
        inpath1="paired.1.fastq",
        inpath2="paired.2.fastq",
        expected1="paired-too-short.1.fastq",
        expected2="paired-too-short.2.fastq",
        aligner=aligner,
        expected_other=(
            f"paired_{aligner}.1.fastq", f"paired_{aligner}.2.fastq"
        )
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_too_short_output_paired_option_missing(run_trimmer, aligner, tmp_path):
    with pytest.raises(SystemExit):
        p1 = tmp_path / "temp-too-short.1.fastq"
        run_trimmer(
            f"-a TTAGACATAT -A CAGTGGAGTA -m 14 --too-short-output {p1}",
            inpath1="paired.1.fastq",
            inpath2="paired.2.fastq",
            expected1="paired.1.fastq",
            expected2="paired.2.fastq",
            aligner=aligner,
        )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_custom_bisulfite_1(run_trimmer, aligner):
    run_trimmer(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 -q 0 --bisulfite 2,2,1,1",
        inpath1="paired_bis_{aligner}.1.fastq",
        inpath2="paired_bis_{aligner}.2.fastq",
        expected1="paired_bis1_{aligner}.1.fastq",
        expected2="paired_bis1_{aligner}.2.fastq",
        aligner=aligner,
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_custom_bisulfite_2(run_trimmer, aligner):
    run_trimmer(
        "-a TTAGACATAT -A CAGTGGAGTA -m 10 -q 0 --bisulfite 20,20,1,1;0,0,0,0",
        inpath1="paired_bis_{aligner}.1.fastq",
        inpath2="paired_bis_{aligner}.2.fastq",
        expected1="paired_bis2_{aligner}.1.fastq",
        expected2="paired_bis2_{aligner}.2.fastq",
        aligner=aligner,
    )


def test_no_insert_match(run_trimmer):
    # with -O
    # Note: this fails if you set -e 0.3 because the higher error rate enables a
    # match of 7 bp, even though it has 2 errors. This illustrates why using
    # --adapter-max-rmp is better.
    # run_trimmer('-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG
    # -A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -O 7 -m 25 -q 0
    # --trim-n',
    #     inpath1='insert.1.fastq', inpath2='insert.2.fastq',
    #     expected1='insert.1.fastq', expected2='insert.2.fastq',
    #     aligner='insert'
    # )
    # with --adapter-max-rmp
    run_trimmer(
        "-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG "
        "-A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT -e 0.3 "
        "--adapter-max-rmp 0.001 -m 25 -q 0 --trim-n",
        inpath1="insert.1.fastq",
        inpath2="insert.2.fastq",
        expected1="insert.1.fastq",
        expected2="insert.2.fastq",
        aligner="insert",
    )


def test_overwrite(run_trimmer):
    run_trimmer(
        "-w 10,30,10",
        inpath1="lowq.fastq",
        inpath2="highq.fastq",
        expected1="lowq.fastq",
        expected2="highq.fastq",
    )


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_no_writer_process(run_trimmer, aligner):
    # TODO: check contents
    output_dir, _, _ = run_trimmer(
        "--threads 3 --no-writer-process --batch-size 1 "
        "-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG "
        "-A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        inpath1="big.1.fq",
        inpath2="big.2.fq",
        expected1="out.1.fastq",
        expected2="out.2.fastq",
        aligner=aligner,
        assert_output_equal=False,
    )
    # TODO: If the final worker doesn't get the chance to process any
    #  batches, the last output file is never created.
    for i in (1, 2):
        for j in (0, 1):
            assert (output_dir / f"tmp{i}-out.{i}.{j}.fastq").exists()


@pytest.mark.parametrize("aligner", BACK_ALIGNERS)
def test_summary(run_trimmer, aligner):
    output_dir, infiles, summary = run_trimmer(
        "--threads 2 "
        "-a AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG "
        "-A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        inpath1="big.1.fq",
        inpath2="big.2.fq",
        expected1="out.1.fastq",
        expected2="out.2.fastq",
        aligner=aligner,
        assert_output_equal=False,
    )

    assert summary is not None
    assert isinstance(summary, dict)
    assert summary["command"] == "trim"
    assert summary["options"]["orig_args"] == (
        "--threads",
        "2",
        "-a",
        "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG",
        "-A",
        "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        "--aligner",
        aligner,
        "-pe1",
        str(infiles[0]),
        "-pe2",
        str(infiles[1]),
        "-o",
        str(output_dir / "tmp1-out.1.fastq"),
        "-p",
        str(output_dir / "tmp2-out.2.fastq"),
    )
    assert summary["sample_id"] == "big"
    assert summary["mode"] == "parallel"
    assert summary["threads"] == 2
    # infiles are 100 125 bp PE reads
    expected_record_counts = {0: 100}
    assert summary["record_counts"] == expected_record_counts
    expected_bp_counts = {0: [12500, 12500]}
    assert summary["bp_counts"] == expected_bp_counts
    assert "timing" in summary
    assert "wallclock" in summary["timing"]
    assert summary["timing"]["wallclock"] > 0
    assert "cpu" in summary["timing"]
    assert summary["timing"]["cpu"] > 0


def test_sam(run_trimmer):
    run_trimmer(
        "-a TTAGACATAT -A CAGTGGAGTA -m 14 --output-format sam",
        "paired_insert.sam",
        "paired.sam",
        interleaved_input=True,
        interleaved_output=True,
        aligner="insert",
    )


# def test_long_reads():
#    run_trimmer(''
#        inpath1='long.1.fq', inpath2='long.2.fq',
#        expected1=, expected2=,
#        aligner=BACK_ALIGNERS)


def test_issue68(run_trimmer):
    run_trimmer(
        "--error-rate 0.20 --insert-match-error-rate 0.30 --minimum-length 20 "
        "--aligner insert -a AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC "
        "-A AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
        inpath1="issue68.1.fq",
        inpath2="issue68.2.fq",
        expected1="issue68.1.fq",
        expected2="issue68.2.fq",
        aligner="insert",
    )



from xphyle import STDOUT

from atropos.commands.console import (
    BaseCommandConsole, add_common_options, validate_common_options
)
from atropos.commands.error import ErrorCommand
from atropos.commands.error.reports import ErrorReportGenerator
from atropos.utils import classproperty
from atropos.utils.argparse import (
    AtroposArgumentParser,
    Namespace,
    writable_file
)


class ErrorCommandConsole(ErrorCommand, ErrorReportGenerator, BaseCommandConsole):
    @classproperty
    def description(cls) -> str:
        return """
        atropos error-se input.fastq
        atropos error -pe in1.fq -pe2 in2.fq
        """

    @classproperty
    def usage(cls) -> str:
        return """
        Estimate the sequencing error rate. Use this command to help determine
        the quality of your data, and to decide the value for the max
        error rate (-e) parameter. Normal error for an Illumina experiment
        is around 1% or less. We recommend setting -e to 10x the empirical
        error rate."""

    @classmethod
    def _add_arguments(cls, parser: AtroposArgumentParser) -> None:
        add_common_options(parser)
        cls._add_error_options(parser)

    @classmethod
    def _validate_options(
        cls, options: Namespace, parser: AtroposArgumentParser
    ) -> None:
        validate_common_options(options, parser)
        cls._validate_error_options(options, parser)

    @classmethod
    def _add_error_options(cls, parser: AtroposArgumentParser):
        parser.set_defaults(
            max_reads=10000, counter_magnitude="K", report_formats=['txt']
        )
        group = parser.add_group("Error Estimation")
        group.add_argument(
            "-a",
            "--algorithm",
            choices=('quality', 'shadow'),
            default="quality",
            help="Method for estimating error rates; quality = base qualities, "
            "shadow = shadow regression. Be advised that the 'shadow' "
            "method is incredibly slow.",
        )
        group.add_argument(
            "-m",
            "--max-bases",
            type=int,
            default=None,
            help="Maximum number of bases to use in the error calculation, "
            "starting from the 5' end of the read.",
        )
        group = parser.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writable_file,
            default=STDOUT,
            help="File in which to write the summary of the estimated error "
            "rates. (stdout)",
        )
        group.add_argument(
            "--output_formats",
            nargs="*",
            choices=cls.list_report_formats(),
            default=None,
            metavar="FORMAT",
            dest="report_formats",
            help="Report type(s) to generate. If multiple, '--output' "
            "is treated as a prefix and the appropriate extensions are "
            "appended. If unspecified, the format is guessed from the "
            "file extension. Supported formats are: txt, json, yaml, "
            "pickle. See the documentation for a full description of "
            "the structured output (json/yaml/pickle formats).",
        )

    @staticmethod
    def _validate_error_options(options: Namespace, parser: AtroposArgumentParser):
        options.report_file = options.output



from argparse import Namespace

import math
import textwrap
from typing import (
    Any, Dict, IO, Optional, Sequence, Tuple, TypeVar, Union, cast,
    Callable,
)

from atropos.commands.reports import (
    DefaultReportGenerator, BaseReportWriter, ReportWriter
)
from atropos.utils import classproperty, truncate_string
from atropos.utils.statistics import weighted_median


INDENT = " " * 2
PARAGRAPH = textwrap.TextWrapper()
INDENTED = textwrap.TextWrapper(initial_indent=INDENT, subsequent_indent=INDENT)


class Printer:
    """
    Manages printing to a file.

    Args:
        outfile: The output file.
        kwargs: Additional keyword arguments passed to the print function.
    """

    def __init__(self, outfile: IO, indent: str = "", **kwargs):
        self.outfile = outfile
        self.indent = indent
        self.print_args = kwargs

    def __call__(self, *args, indent: Union[int, str] = None, **kwargs):
        if isinstance(indent, int):
            indent_str = self.indent * cast(int, indent)
        else:
            indent_str = cast(str, indent) or self.indent

        if indent_str:
            self._print(indent_str, end="")

        self._print(*args, **kwargs)

    def _print(self, *args, **kwargs):
        if self.print_args:
            print_args = self.print_args.copy()
            print_args.update(kwargs)
        else:
            print_args = kwargs

        print(*args, file=self.outfile, **print_args)

    def newline(self):
        """
        Prints a newline.
        """
        print(file=self.outfile)


class TitlePrinter(Printer):
    """
    Printer that formats titles.

    Args:
        outfile: The output file.
        levels: The formatting associated with different header levels.
        kwargs: Additional keyword arguments passed to the print function.
    """

    def __init__(
        self,
        outfile: IO,
        levels: Sequence[Tuple[str, Optional[str]]] = (
            ("=", "="),
            ("-", "-"),
            ("-", None),
            ("~", None),
        ),
        **kwargs,
    ):
        super().__init__(outfile, **kwargs)
        self.levels = levels

    def __call__(
        self, *title: str, level: Optional[int] = None, newline: bool = True, **kwargs
    ):
        title = " ".join(title)
        underline = None
        width = 0

        if level is not None:
            if level >= len(self.levels):
                raise ValueError(f"Invalid level: {level}")

            underline, overline = self.levels[level]
            if overline is True:
                overline = underline

            width = len(title)

            if overline:
                self._print(overline * width, **kwargs)

        self._print(title, **kwargs)

        if level is not None and underline:
            self._print(underline * width, **kwargs)

        if newline:
            self.newline()


class RowPrinter(Printer):
    """
    Printer that formats rows in a table.
    """

    def __init__(
        self,
        outfile: IO,
        colwidths: Union[int, Tuple[int, ...]] = 10,
        justification: Union[str, Tuple[str, ...]] = ("<", ">"),
        indent: Union[str, Tuple[str, ...]] = "",
        pct: bool = False,
        default: int = 0,
        **kwargs,
    ):
        """
        Args:
            outfile: The output file.
            colwidths: Column widths.
            justification: Column justifiations ('<' for left, '>' for right).
            indent: Column indents.
            pct: Whether floats should be formatted as percentages.
            default: Default value for None's.
            kwargs: Additional keyword arguments passed to the print function.

        colwidths, justification, and indent can be longer or shorter than the
        number of arguments; if shorter, the last value in the list is repeated;
        if longer, the list is truncated.
        """
        super().__init__(outfile, **kwargs)
        self.colwidths, self.justification, self.indent = (
            (arg,) if isinstance(arg, typ) else tuple(arg)
            for arg, typ in zip((colwidths, justification, indent), (int, str, str))
        )
        self.pct = pct
        self.default = default

    def print_rows(
        self,
        *rows,
        header: Optional[Sequence[Union[str, Tuple[str, str]]]] = None,
        **kwargs,
    ):
        """
        Prints multiple rows. Automatically figures out column widths.

        Args:
            rows: Rows to print.
            header: Header row.
            kwargs: Additional keyword arguments to self.__call__.
        """
        colwidths = tuple(sizeof(*x) for x in zip(*rows))

        if header:
            if isinstance(header[0], str):
                header_widths = (sizeof(h) for h in header)
                header_rows = [header]
            else:
                header_widths = (
                    max(sizeof(h) for h in header_part) for header_part in header
                )
                header_rows = list(zip(*header))

            colwidths = tuple(max(h, c) for h, c in zip(header_widths, colwidths))

            for i, header_row in enumerate(header_rows, 1):
                self(
                    *header_row,
                    colwidths=colwidths,
                    header=(i == len(header_rows)),
                    **kwargs,
                )

        for row in rows:
            self(*row, colwidths=colwidths)

    def __call__(
        self,
        *args,
        colwidths: Optional[Tuple[int, ...]] = None,
        extra_width: Optional[int] = None,
        justification: Optional[Tuple[str, ...]] = None,
        extra_justification: Optional[str] = None,
        indent: Optional[Tuple[str, ...]] = None,
        extra_indent: Optional[str] = None,
        header: bool = False,
        underline: str = "-",
        pct: Optional[bool] = None,
        default: Optional[Any] = None,
        **kwargs,
    ):
        """
        Prints a row.

        Args:
            args: Fields in the row.
            colwidths: Row-specific colwidths.
            justification: Row-specific justifications.
            indent: Row-specific indents.
            extra_width: colwidth to use for extra fields.
            extra_justification: Justification to use for extra fields.
            extra_indent: Indent to use for extra fields.
            header: Whether this is a header row.
            underline: Whether to use an underline after the header row. Either
                a bool or a character.
            pct: Whether floating point values should be formatted as
                percentages.
            default: Default value.
            kwargs: Additional keyword arguments to pass to print.
        """
        ncols = len(args)
        if ncols == 0:
            self.newline()
            return

        if pct is None:
            pct = self.pct

        T = TypeVar("T")

        def adjust(tup: Tuple[T, ...], extra: Optional[T] = None) -> Tuple[T, ...]:
            """
            Adjust a tuple. If longer than the number of columns, truncate; if
            shorter, fill in by repeating the last element.
            """
            tlen = len(tup)
            if tlen == ncols:
                return tup
            elif tlen > ncols:
                return tup[:ncols]
            else:
                return tup + ((extra or tup[-1],) * (ncols - tlen))

        colwidths, justification, indent = (
            adjust(arr, extra)
            for arr, extra in zip(
                (
                    colwidths or self.colwidths,
                    justification or self.justification,
                    indent or self.indent,
                ),
                (extra_width, extra_justification, extra_indent),
            )
        )

        # adjust colwidths if this is a header
        if header:
            colwidths = tuple(max(w, len(str(a))) for w, a in zip(colwidths, args))

        fmt_str = []
        fmt_args = []

        for i, (value, width, just, ind) in enumerate(
            zip(args, colwidths, justification, indent)
        ):
            if value is None:
                value = default or self.default

            if isinstance(value, str):
                typ = "s"
                if len(value) > width:
                    value = truncate_string(value, width)
            elif isinstance(value, float):
                typ = ",.1" + ("%" if pct else "f")
            else:
                typ = ",d"

            fmt_str.append(
                ind + "{" + str(i) + ":" + just + str(width - len(ind)) + typ + "}"
            )
            fmt_args.append(value)

        self._print(" ".join(fmt_str).format(*fmt_args), **kwargs)

        if header:
            sepline = " ".join((underline * width) for width in colwidths)
            self._print(sepline, **kwargs)


class LegacyTextReportWriter(BaseReportWriter):
    @classproperty
    def extensions(cls) -> Sequence[str]:
        return "txt",

    def __init__(self, options: Namespace):
        super().__init__(options)

    def serialize(self, summary: dict, stream: IO) -> None:
        self._print_summary_report(summary, stream)
        if "trim" in summary:
            self._print_trim_report(summary, stream)
        if "pre" in summary:
            self._print_pre_trim_report(summary, stream)
        if "post" in summary:
            self._print_post_trim_report(summary, stream)

    @classmethod
    def _print_summary_report(cls, summary: dict, stream: IO):
        """
        Prints the top-level summary report.

        Args:
            summary: The summary dict.
            stream: The output file object.
        """
        _print_title = TitlePrinter(stream)
        _print = Printer(stream)
        _print_title("Atropos", level=0)
        _print(f"Atropos version: {summary['version']}")
        _print(f"Python version: {summary['python']}")
        _print(
            f"Command line parameters: {summary['command']} "
            f"{' '.join(summary['options']['orig_args'])}"
        )
        _print()
        _print(f"Sample ID: {summary['sample_id']}")
        _print(f"Input format: {summary['derived']['input_format']}")
        _print("Input files:")

        for infile in summary["input"]["input_names"]:
            if infile is not None:
                _print(infile, indent=INDENT)

        _print()

        timing = summary["timing"]
        total = summary["total_record_count"]
        wctime = [f"Wallclock time: {timing['wallclock']:.2F} s"]

        if total > 0:
            wctime.append(
                f"({1e6 * timing['wallclock'] / total:.0F} us/read; "
                f"{total / timing['wallclock'] * 60 / 1e6:.2F} M reads/minute)"
            )

        _print(f"Start time: {timing['start']}")
        _print(*wctime)
        _print(f"CPU time (main process): {timing['cpu']:.2F} s")
        _print()

    @classmethod
    def _print_trim_report(cls, summary: dict, stream: IO):
        """
        Prints the trimming report.

        Args:
            summary: Summary dict.
            stream: Open output stream.
        """
        total_bp = sum(summary["total_bp_counts"])
        max_width = len(str(total_bp))
        # account for commas
        max_width += max_width // 3
        _print = RowPrinter(stream, (35, max_width))
        total = summary["total_record_count"]

        if total == 0:
            _print(
                "No reads processed! Either your input file is empty or you "
                "used the wrong -f/--format parameter."
            )
            return

        modifiers, filters, formatters = (
            summary["trim"][key] for key in ("modifiers", "filters", "formatters")
        )

        adapter_cutter = None
        error_corrector = None

        for modifier_dict in modifiers.values():
            if adapter_cutter is None and "adapters" in modifier_dict:
                adapter_cutter = modifier_dict
                break

            if error_corrector is None and "bp_corrected" in modifier_dict:
                error_corrector = modifier_dict

        correction_enabled = summary["options"]["correct_mismatches"]
        corrected = None
        trimmers = []

        for name, mod in modifiers.items():
            if "bp_trimmed" in mod:
                trimmers.append((name, mod))

            if correction_enabled and "records_corrected" in mod:
                corrected = mod

        _print_title = TitlePrinter(stream)
        paired = summary["options"]["paired"]
        pairs_or_reads = "Pairs" if paired else "Reads"

        _print_title("Trimming", level=1)
        _print(pairs_or_reads, "records", "fraction", header=True)
        _print(f"Total {'read pairs' if paired else 'reads'} processed:", total)

        if adapter_cutter:
            if paired:
                for read in range(2):
                    _print(
                        f"Read {read + 1} with adapter:",
                        adapter_cutter["records_with_adapters"][read],
                        adapter_cutter["fraction_records_with_adapters"][read],
                        indent=(INDENT, ""),
                        pct=True,
                    )
            else:
                _print(
                    "Reads with adapters:",
                    adapter_cutter["records_with_adapters"][0],
                    adapter_cutter["fraction_records_with_adapters"][0],
                    pct=True,
                )

        def _print_filter(_name, sep):
            if _name in filters:
                _print(
                    f"{pairs_or_reads} {sep} {_name.replace('_', ' ')}:",
                    filters[_name]["records_filtered"],
                    filters[_name]["fraction_records_filtered"],
                    pct=True,
                )

        _print_filter("too_short", "that were")
        _print_filter("too_long", "that were")
        _print_filter("too_many_n", "with")
        _print(
            f"{pairs_or_reads} written (passing filters):",
            formatters["records_written"],
            formatters["fraction_records_written"],
            pct=True,
        )

        if corrected:
            _print(
                "Pairs corrected:",
                corrected["records_corrected"],
                corrected["fraction_records_corrected"],
                pct=True,
            )

        _print()
        _print("Base pairs", "bp", "fraction", header=True)
        _print("Total bp processed:", total_bp)

        if paired:
            for read in range(2):
                _print(
                    f"Read {read + 1}:",
                    summary["total_bp_counts"][read],
                    indent=(INDENT, ""),
                )

        def _print_bp(title, data, key, default=0):
            if paired:
                _print(
                    title,
                    data[f"total_{key}"],
                    data[f"fraction_total_{key}"],
                    pct=True,
                )

                for _read in range(2):
                    _print(
                        f"Read {_read + 1}:",
                        data[key][_read],
                        data[f"fraction_{key}"][_read],
                        indent=(INDENT, ""),
                        pct=True,
                        default=default,
                    )
            else:
                _print(
                    title,
                    data[key][0],
                    data[f"fraction_{key}"][0],
                    pct=True,
                    default=default,
                )

        for name, mod in trimmers:
            _print_bp(mod["desc"], mod, "bp_trimmed")

        _print_bp("Total bp written (filtered):", formatters, "bp_written")

        if error_corrector:
            _print_bp("Total bp corrected:", error_corrector, "bp_corrected")

        if adapter_cutter:
            _print()
            adapters = adapter_cutter["adapters"]
            cls._print_adapter_report(adapters, stream, paired, total, max_width)

    @classmethod
    def _print_adapter_report(
        cls,
        adapters: Sequence[dict],
        stream: IO,
        paired: bool,
        total_records: int,
        max_width: int,
    ):
        """
        Prints details for a adapters.

        Args:
            adapters: Sequence of adapter info dicts.
            stream: The output file
            paired: Whether the data is paired-end.

            max_width: Max column width.
        """
        adapter_lenghts = []

        for pair in adapters:
            if pair:
                for adapter in pair.values():
                    if adapter["where"]["name"] == "linked":
                        adapter_lenghts.append(
                            3
                            + len(adapter["front_sequence"] + adapter["back_sequence"])
                        )
                    else:
                        adapter_lenghts.append(len(adapter["sequence"]))

        max_seq_len = max(adapter_lenghts)
        _print = Printer(stream)
        _print_title = TitlePrinter(stream)
        _print_adj = RowPrinter(stream, (12, 5), pct=True, indent=(INDENT, ""))
        seq_printer = RowPrinter(
            stream, (max_seq_len, 14, 3, max_width), ("<", "<", ">")
        )
        hist_printer = RowPrinter(stream, justification=(">", ">", ">", ">", "<"))

        def print_error_ranges(adapter_length: int, error_rate: float):
            """
            Prints max number of errors for ranges of adapter match lengths.
            """
            _print("No. of allowed errors:")

            prev = 0

            for errors in range(1, int(error_rate * adapter_length) + 1):
                range_start = int(errors / error_rate)
                _print(f"{prev}-{range_start - 1} bp: {errors - 1};", end=" ")
                prev = range_start

            if prev == adapter_length:
                _print(f"{adapter_length} bp: {int(error_rate * adapter_length)}")
            else:
                _print("{prev}-{adapter_length} bp: {int(error_rate * adapter_length)}")

            _print()

        def print_histogram(
            data: dict,
            adapter_length: int,
            num_reads: int,
            error_rate: float,
            errors: dict,
            match_probabilities: Sequence[float],
        ):
            """
            Prints a histogram. Also, print the no. of reads expected to be
            trimmed by chance (assuming a uniform distribution of nucleotides in
            the reads).

            Args:
                data: A dictionary mapping lengths of trimmed sequences to their
                    respective frequency
                adapter_length: adapter length
                num_reads: total no. of reads.
                error_rate: Max error rate.
                errors: Histogram of actual numbers of errors.
                match_probabilities: List of random match proabilities for each
                    position in the adapter seqeunce.
            """
            hist = []
            hist_errors = []

            for length, count in data.items():
                # when length surpasses adapter_length, the probability does not
                # increase anymore
                estimated = num_reads * match_probabilities[min(length, adapter_length)]
                hist.append(
                    [
                        length,
                        count,
                        estimated,
                        int(error_rate * min(length, adapter_length)),
                    ]
                )
                hist_errors.append(errors["rows"][length])

            col_sizes = [len(str(max(col))) for col in zip(*hist_errors)]

            def _format_hist_errors(_errs):
                trailing = True
                hist_str = []

                for e_ctr, e in reversed(list(enumerate(_errs))):
                    if trailing and e == 0:
                        pass
                    else:
                        trailing = False
                        hist_str.append(
                            ("{:<" + str(col_sizes[e_ctr]) + "d}").format(e)
                        )

                return " ".join(reversed(hist_str))

            for i, errs in enumerate(hist_errors):
                hist[i].append(_format_hist_errors(errs))

            error_header = " ".join(
                ("{:<" + str(e) + "d}").format(i) for i, e in enumerate(col_sizes)
            )

            hist_printer.print_rows(
                *hist,
                header=(
                    ("length", ""),
                    ("count", ""),
                    ("expect", ""),
                    ("max.err", ""),
                    ("error counts", error_header),
                ),
            )
            hist_printer.newline()

        def print_adjacent_bases(bases: Dict[str, int]):
            """
            Prints a summary of the bases preceding removed adapter sequences.
            Print a warning if one of the bases is overrepresented and there are
            at least 20 preceding bases available.

            Return:
                True if a warning was printed.
            """
            total = sum(bases.values())
            if total == 0:
                return False

            _print("Bases preceding removed adapters:")

            warnbase = None

            for base in ["A", "C", "G", "T", ""]:
                base_label = base if base != "" else "none/other"
                fraction = bases[base] / total
                _print_adj(base_label, fraction)
                if fraction > 0.8 and base != "":
                    warnbase = base_label

            if total >= 20 and warnbase is not None:
                _print("WARNING:")
                _print(
                    "\n".join(
                        INDENTED.wrap(
                            f"The adapter is preceded by '{warnbase}'extremely often. "
                            f"The provided adapter sequence may be incomplete. To fix "
                            f"the problem, add '{warnbase}' to the beginning of the "
                            f"adapter sequence."
                        )
                    )
                )
                _print()
                return True

            _print()
            return False

        warning = False

        for pair in range(2 if paired else 1):
            if adapters[pair] is None:
                continue

            header = "Adapter {}"

            if paired:
                header = ("First read: " if pair == 0 else "Second read: ") + header

            for name, adapter in adapters[pair].items():
                _print_title(header.format(name), level=1)

                where_name = adapter["where"]["name"]
                front_len = 0
                back_len = 0
                seq_len = 0

                if where_name == "linked":
                    front_len, back_len = [
                        len(adapter[s]) for s in ("front_sequence", "back_sequence")
                    ]
                    seq_printer.print_rows(
                        (
                            f"{adapter['front_sequence']}...{adapter['back_sequence']}",
                            "linked",
                            f"{front_len}+{back_len}",
                            adapter["total_front"],
                            adapter["total_back"],
                        ),
                        header=(
                            "Sequence",
                            "Type",
                            "Length",
                            "Trimmed (x)",
                            "Half matches (x)",
                        ),
                    )
                else:
                    seq_len = len(adapter["sequence"])
                    seq_printer.print_rows(
                        (
                            adapter["sequence"],
                            adapter["where"]["desc"],
                            seq_len,
                            adapter["total"],
                        ),
                        header=("Sequence", "Type", "Length", "Trimmed (x)"),
                    )

                _print()

                if adapter["total"] == 0:
                    continue

                if where_name == "anywhere":
                    _print(
                        adapter["total_front"],
                        "times, it overlapped the 5' end of a read",
                    )
                    _print(
                        adapter["total_back"],
                        "times, it overlapped the 3' end or was within the read",
                    )
                    _print()
                    print_error_ranges(seq_len, adapter["max_error_rate"])
                    _print("Overview of removed sequences (5'):")
                    print_histogram(
                        adapter["lengths_front"],
                        seq_len,
                        total_records,
                        adapter["max_error_rate"],
                        adapter["errors_front"],
                        adapter["match_probabilities"],
                    )
                    _print()
                    _print("Overview of removed sequences (3' or within):")
                    print_histogram(
                        adapter["lengths_back"],
                        seq_len,
                        total_records,
                        adapter["max_error_rate"],
                        adapter["errors_back"],
                        adapter["match_probabilities"],
                    )
                elif where_name == "linked":
                    print_error_ranges(front_len, adapter["front_max_error_rate"])
                    print_error_ranges(back_len, adapter["back_max_error_rate"])
                    _print("Overview of removed sequences at 5' end:")
                    print_histogram(
                        adapter["front_lengths_front"],
                        front_len,
                        total_records,
                        adapter["front_max_error_rate"],
                        adapter["front_errors_front"],
                        adapter["front_match_probabilities"],
                    )
                    _print()
                    _print("Overview of removed sequences at 3' end:")
                    print_histogram(
                        adapter["back_lengths_back"],
                        back_len,
                        total_records,
                        adapter["back_max_error_rate"],
                        adapter["back_errors_back"],
                        adapter["back_match_probabilities"],
                    )
                elif where_name in ("front", "prefix"):
                    print_error_ranges(seq_len, adapter["max_error_rate"])
                    _print("Overview of removed sequences:")
                    print_histogram(
                        adapter["lengths_front"],
                        seq_len,
                        total_records,
                        adapter["max_error_rate"],
                        adapter["errors_front"],
                        adapter["match_probabilities"],
                    )
                elif where_name in ("back", "suffix"):
                    print_error_ranges(seq_len, adapter["max_error_rate"])
                    warning = warning or print_adjacent_bases(adapter["adjacent_bases"])
                    _print("Overview of removed sequences:")
                    print_histogram(
                        adapter["lengths_back"],
                        seq_len,
                        total_records,
                        adapter["max_error_rate"],
                        adapter["errors_back"],
                        adapter["match_probabilities"],
                    )

        if warning:
            _print("WARNING:")
            _print(
                "\n".join(
                    INDENTED.wrap(
                        "One or more of your adapter sequences may be incomplete. "
                        "Please see the detailed output above."
                    )
                )
            )

    @classmethod
    def _print_pre_trim_report(cls, summary: dict, stream: IO):
        """
        Prints pre-trimming metrics.

        Args:
            summary: The summary dict.
            stream: The output file.
        """
        pre = summary["pre"]
        _print_title = TitlePrinter(stream)
        _print = Printer(stream)
        _print_title("Pre-trimming metrics", level=1)

        for source, data in pre.items():
            _print_title("Source", level=3, newline=False)

            # TODO: When multi-file input is supported, this code will need to
            #  get summary['input']['input_names'][source]
            for read, src in enumerate(summary["input"]["input_names"], 1):
                if src is None:
                    continue

                _print(f"Read {read}: {src}")

            _print()

            cls._print_metrics_report(data, stream)

    @classmethod
    def _print_post_trim_report(cls, summary: dict, outfile: IO):
        """
        Prints post-trimming metrics.

        Args:
            summary: The summary dict.
            outfile: The output file.
        """
        post = summary["post"]
        _print_title = TitlePrinter(outfile)
        _print = Printer(outfile)
        _print_title("Post-trimming metrics", level=1)

        for dest, metrics in post.items():
            _print_title(f"Destination: {dest}", level=2)

            for source, data in metrics.items():
                _print_title("Source", level=3, newline=False)

                # TODO: When multi-file input is supported, this code will need to
                #  get summary['input']['input_names'][source]

                for read, src in enumerate(summary["input"]["input_names"], 1):
                    if src is None:
                        continue

                    _print(f"Read {read}: {src}")

                _print()

                cls._print_metrics_report(data, outfile)

    @classmethod
    def _print_metrics_report(cls, data, outfile):
        """
        Prints metrics.

        Args:
            data: The metrics dict.
            outfile: The output file.
        """
        paired = "read2" in data
        max_count = data["read1"]["counts"]

        if paired:
            max_count = max(max_count, data["read2"]["counts"])

        max_width = len(str(max_count))
        # add space for commas and column separation
        max_width += (max_width // 3) + 1
        _print_title = TitlePrinter(outfile)
        _print = RowPrinter(outfile, (35, max_width))

        def _print_histogram(title, hist1, hist2=None):
            _print_title(title, level=2)

            if hist1 is None:
                _print("No Data")
                return

            if hist2:
                hist = (
                    (key, hist1.get(key, 0), hist2.get(key, 0))
                    for key in sorted(set(hist1.keys()) | set(hist2.keys()))
                )
            else:
                hist = sorted(hist1.items(), key=lambda x: x[0])

            for histbin in hist:
                _print(*histbin)

        def _print_base_histogram(title, hist, extra_width=4, index_name="Pos"):
            _print_title(title, level=2)

            if hist is None:
                _print("No Data")
                return

            _print(index_name, *hist["columns"], header=True, extra_width=extra_width)

            for pos, row in hist["rows"].items():
                total_count = sum(row)
                base_pcts = (round(count * 100 / total_count, 1) for count in row)
                _print(pos, *base_pcts, extra_width=extra_width)

        def _print_tile_histogram(title, hist):
            if hist is None:
                _print_title(title, level=2)
                _print("No Data")
                return

            ncol = len(hist["columns"])
            max_tile_width = (
                max(4, len(str(math.ceil(data["read1"]["counts"] / ncol)))) + 1
            )
            _print_base_histogram(
                title, hist, extra_width=max_tile_width, index_name="Tile"
            )

        def _print_tile_base_histogram(title: str, hist: dict):
            """Print a histogram of position x tile, with values as the median
            base quality.
            """
            _print_title(title, level=2)
            if hist is None:
                _print("No Data")
                return

            quals = hist["columns"]
            tiles = hist["columns2"]
            ncol = len(tiles)
            max_tile_width = (
                max(4, len(str(math.ceil(data["read1"]["counts"] / ncol)))) + 1
            )
            _print("Pos", *tiles, header=True, extra_width=max_tile_width)

            for pos, tiles in hist["rows"].items():
                # compute the weighted median for each tile at each position
                _print(
                    pos,
                    *(
                        weighted_median(quals, tile_counts)
                        for tile_counts in tiles.values()
                    ),
                    extra_width=max_tile_width,
                )

        _print("", "Read1", "Read2", header=True)

        # Sequence-level metrics
        _print(
            "Read pairs:" if paired else "Reads:",
            data["read1"]["counts"],
            data["read2"]["counts"],
        )
        _print()
        _print_histogram(
            "Sequence lengths:",
            data["read1"]["lengths"]["hist"],
            data["read2"]["lengths"]["hist"],
        )
        _print()
        if "qualities" in data["read1"]:
            _print_histogram(
                "Sequence qualities:",
                data["read1"]["qualities"]["hist"],
                data["read2"]["qualities"]["hist"],
            )
            _print()
        _print_histogram(
            "Sequence GC content (%)",
            data["read1"]["gc"]["hist"],
            data["read2"]["gc"]["hist"],
        )
        _print()
        if "tile_sequence_qualities" in data["read1"]:
            _print_tile_histogram(
                "Read 1 per-tile sequence qualities (%)",
                data["read1"]["tile_sequence_qualities"],
            )
            _print()
            _print_tile_histogram(
                "Read 2 per-tile sequence qualities (%)",
                data["read2"]["tile_sequence_qualities"],
            )
            _print()

        # Base-level metrics
        if "base_qualities" in data["read1"]:
            _print_base_histogram(
                "Read 1 base qualities (%)", data["read1"]["base_qualities"]
            )
            _print()
            _print_base_histogram(
                "Read 2 base qualities (%)", data["read2"]["base_qualities"]
            )
            _print()
        _print_base_histogram("Read 1 base composition (%)", data["read1"]["bases"])
        _print()
        _print_base_histogram("Read 2 base composition (%)", data["read2"]["bases"])
        _print()

        if "tile_base_qualities" in data["read1"]:
            _print_tile_base_histogram(
                "Read 1 per-tile base qualities (%)",
                data["read1"]["tile_base_qualities"],
            )
            _print()
            _print_tile_base_histogram(
                "Read 2 per-tile base qualities (%)",
                data["read2"]["tile_base_qualities"],
            )
            _print()


def sizeof(*x: Union[str, int, float], seps: bool = True, prec: int = 1):
    """
    Returns the largest string size of all objects in x, where x is a sequence of
    string or numeric values.

    Args:
        *x: The objects to test.
        seps: Whether to include separators (,.) in the size.
        prec: Precision of float values.
    """
    if isinstance(x[0], str):
        return max(len(s) for s in x)
    else:
        if isinstance(x[0], int):
            numlen = len(str(max(x)))
            if seps:
                numlen += numlen // 3
        elif isinstance(x[0], float):
            numlen = len(str(round(max(x), prec)))
            if seps:
                numlen += (numlen - prec - 1) // 3
        else:
            raise ValueError(f"Unexpected data type: {x[0].__class__}")

        return numlen


class LegacyReportGenerator(DefaultReportGenerator):
    @classproperty
    def report_formats(cls) -> Dict[str, Callable[[Namespace], ReportWriter]]:
        report_formats = super().report_formats
        report_formats["txt"] = LegacyTextReportWriter
        return report_formats



# TODO:
#  test with the --output option
#  test reading from standard input
import pytest
from xphyle import open_

from atropos.console import execute_cli
from atropos.utils import LOGGING_CONFIG, ReturnCode, no_import

from .utils import no_internet

try:
    import lzma
except ImportError:
    lzma = None


def test_example(run_trimmer):
    run_trimmer("-N -b ADAPTER", "example.fa", "example.fa")


def test_example_stdout(run_trimmer):
    run_trimmer("-N -b ADAPTER", "example.fa", "example.fa", stdout=True)


def test_small(run_trimmer):
    run_trimmer("-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq")


def test_empty(run_trimmer):
    run_trimmer("-a TTAGACATATCTCCGTCG", "empty.fastq", "empty.fastq")


def test_newlines(run_trimmer):
    run_trimmer("-e 0.12 -b TTAGACATATCTCCGTCG", "dos.fastq", "dos.fastq")


def test_lowercase(run_trimmer):
    run_trimmer("-b ttagacatatctccgtcg", "lowercase.fastq", "small.fastq")


def test_rest(run_trimmer, tmp_path):
    run_trimmer(
        ["-b", "ADAPTER", "-N", "-r", "rest.txt"],
        "rest.fa",
        "rest.fa",
        expected_other=["rest.txt"],
    )


def test_restfront(run_trimmer, tmp_path):
    run_trimmer(
        ["-g", "ADAPTER", "-N", "-r", "restfront.txt"],
        "restfront.fa",
        "rest.fa",
        expected_other=["restfront.txt"],
    )


def test_discard_trimmed(run_trimmer):
    run_trimmer(
        "-b TTAGACATATCTCCGTCG --discard-trimmed", "discard.fastq", "small.fastq"
    )


def test_discard_untrimmed(run_trimmer):
    run_trimmer(
        "-b CAAGAT --discard-untrimmed", "discard-untrimmed.fastq", "small.fastq"
    )


def test_plus(run_trimmer):
    run_trimmer("-e 0.12 -b TTAGACATATCTCCGTCG", "plus.fastq", "plus.fastq")


def test_extensiontxtgz(run_trimmer):
    run_trimmer("-b TTAGACATATCTCCGTCG", "s_1_sequence.txt", "s_1_sequence.txt.gz")


def test_format(run_trimmer):
    run_trimmer("-f fastq -b TTAGACATATCTCCGTCG", "small.fastq", "small.myownextension")


def test_minimum_length(run_trimmer):
    run_trimmer("-c -m 5 -a 330201030313112312", "minlen.fa", "lengths.fa")


def test_too_short(run_trimmer):
    run_trimmer(
        "-c -m 5 -a 330201030313112312 --too-short-output tooshort.fa",
        "minlen.fa",
        "lengths.fa",
        expected_other=["tooshort.fa"],
    )


def test_too_short_no_primer(run_trimmer):
    run_trimmer(
        "-c -m 5 -a 330201030313112312 --trim-primer "
        "--too-short-output tooshort.noprimer.fa",
        "minlen.noprimer.fa",
        "lengths.fa",
        expected_other=["tooshort.noprimer.fa"],
    )


def test_maximum_length(run_trimmer):
    run_trimmer("-c -M 5 -a 330201030313112312", "maxlen.fa", "lengths.fa")


def test_too_long(run_trimmer):
    run_trimmer(
        "-c -M 5 --too-long-output toolong.fa -a 330201030313112312",
        "maxlen.fa",
        "lengths.fa",
        expected_other=["toolong.fa"],
    )


def test_length_tag(run_trimmer):
    run_trimmer(
        "-n 3 -e 0.1 --length-tag length= "
        "-b TGAGACACGCAACAGGGGAAAGGCAAGGCACACAGGGGATAGG "
        "-b TCCATCTCATCCCTGCGTGTCCCATCTGTTCCCTCCCTGTCTCA",
        "454.fa",
        "454.fa",
    )


def test_overlap_a(run_trimmer):
    run_trimmer("-O 10 -a 330201030313112312 -e 0.0 -N", "overlapa.fa", "overlapa.fa")


def test_overlap_b(run_trimmer):
    run_trimmer("-O 10 -b TTAGACATATCTCCGTCG -N", "overlapb.fa", "overlapb.fa")


def test_qualtrim(run_trimmer):
    run_trimmer("-q 10 -a XXXXXX", "lowqual.fastq", "lowqual.fastq")


def test_qualbase(run_trimmer):
    run_trimmer(
        "-q 10 --quality-base 64 -a XXXXXX", "illumina64.fastq", "illumina64.fastq"
    )


def test_quality_trim_only(run_trimmer):
    run_trimmer("-q 10 --quality-base 64", "illumina64.fastq", "illumina64.fastq")


def test_twoadapters(run_trimmer):
    run_trimmer(
        "-a AATTTCAGGAATT -a GTTCTCTAGTTCT", "twoadapters.fasta", "twoadapters.fasta"
    )


def test_polya(run_trimmer):
    run_trimmer(
        "-m 24 -O 10 -a AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        "polya.fasta",
        "polya.fasta",
    )


def test_polya_brace_notation(run_trimmer):
    run_trimmer("-m 24 -O 10 -a A{35}", "polya.fasta", "polya.fasta")


def test_mask_adapter(run_trimmer):
    run_trimmer(
        "-b CAAG -n 3 --mask-adapter", "anywhere_repeat.fastq", "anywhere_repeat.fastq"
    )


def test_gz_multiblock(run_trimmer):
    run_trimmer("-b TTAGACATATCTCCGTCG", "small.fastq", "multiblock.fastq.gz")


def test_suffix(run_trimmer):
    run_trimmer(
        "-c -e 0.12 -a 1=330201030313112312 -y _my_suffix_{name} --strip-f3",
        "suffix.fastq",
        "solid.csfasta",
        qualfile="solid.qual",
    )


def test_read_wildcard(run_trimmer):
    run_trimmer("--match-read-wildcards -b ACGTACGT", "wildcard.fa", "wildcard.fa")


def test_adapter_wildcard(run_trimmer, tmp_path):
    for adapter_type, expected in (
        ("-a", "wildcard_adapter.fa"),
        ("-b", "wildcard_adapter_anywhere.fa"),
    ):
        wildcardtmp = tmp_path / "wildcardtmp.txt"

        run_trimmer(
            ("--wildcard-file", wildcardtmp, adapter_type, "ACGTNNNACGT"),
            expected,
            "wildcard_adapter.fa",
        )

        with open_(wildcardtmp) as wct:
            lines = [line.strip() for line in wct.readlines()]

        assert lines == ["AAA 1", "GGG 2", "CCC 3b", "TTT 4b"]


def test_wildcard_n(run_trimmer):
    run_trimmer(
        "-e 0 -a GGGGGGG --match-read-wildcards", "wildcardN.fa", "wildcardN.fa"
    )


def test_illumina_adapter_wildcard(run_trimmer):
    run_trimmer(
        "-a VCCGAMCYUCKHRKDCUBBCNUWNSGHCGU", "illumina.fastq", "illumina.fastq.gz"
    )


def test_adapter_front(run_trimmer):
    run_trimmer("--front ADAPTER -N", "examplefront.fa", "example.fa")


def test_literal_n(run_trimmer):
    run_trimmer("-N -e 0.2 -a NNNNNNNNNNNNNN", "trimN3.fasta", "trimN3.fasta")


def test_literal_n2(run_trimmer):
    run_trimmer("-N -O 1 -g NNNNNNNNNNNNNN", "trimN5.fasta", "trimN5.fasta")


def test_literal_n_brace_notation(run_trimmer):
    run_trimmer("-N -e 0.2 -a N{14}", "trimN3.fasta", "trimN3.fasta")


def test_literal_n2_brace_notation(run_trimmer):
    run_trimmer("-N -O 1 -g N{14}", "trimN5.fasta", "trimN5.fasta")


def test_anchored_front(run_trimmer):
    run_trimmer("-g ^FRONTADAPT -N", "anchored.fasta", "anchored.fasta")


def test_anchored_front_ellipsis_notation(run_trimmer):
    run_trimmer("-a FRONTADAPT... -N", "anchored.fasta", "anchored.fasta")


def test_anchored_back(run_trimmer):
    run_trimmer("-a BACKADAPTER$ -N", "anchored-back.fasta", "anchored-back.fasta")


def test_anchored_back_no_indels(run_trimmer):
    run_trimmer(
        "-a BACKADAPTER$ -N --no-indels", "anchored-back.fasta", "anchored-back.fasta"
    )


def test_no_indels(run_trimmer):
    run_trimmer(
        "-a TTAGACATAT -g GAGATTGCCA --no-indels", "no_indels.fasta", "no_indels.fasta"
    )


def test_issue_46(run_trimmer, tmp_path):
    wildcardtmp = tmp_path / "wildcardtmp.txt"
    run_trimmer(
        f"--anywhere=AACGTN --wildcard-file={wildcardtmp}",
        "issue46.fasta",
        "issue46.fasta",
    )


def test_strip_suffix(run_trimmer):
    run_trimmer("--strip-suffix _sequence -a XXXXXXX", "stripped.fasta", "simple.fasta")


def test_info_file(run_trimmer):
    run_trimmer(
        [
            "--info-file",
            "illumina.info.txt",
            "-a",
            "adapt=GCCGAACTTCTTAGACTGCCTTAAGGACGT",
        ],
        "illumina.fastq",
        "illumina.fastq.gz",
        expected_other=["illumina.info.txt"],
    )


def test_info_file_times(run_trimmer):
    run_trimmer(
        [
            "--info-file",
            "illumina5.info.txt",
            "--times",
            "2",
            "-a",
            "adapt=GCCGAACTTCTTA",
            "-a",
            "adapt2=GACTGCCTTAAGGACGT",
        ],
        "illumina5.fastq",
        "illumina5.fastq",
        expected_other=["illumina5.info.txt"],
    )


def test_info_file_fasta(run_trimmer, tmp_path):
    infotmp = tmp_path / "infotmp.txt"
    # Just make sure that it runs
    run_trimmer(
        [
            "--info-file",
            infotmp,
            "-a",
            "TTAGACATAT",
            "-g",
            "GAGATTGCCA",
            "--no-indels",
        ],
        "no_indels.fasta",
        "no_indels.fasta",
    )


def test_named_adapter(run_trimmer):
    run_trimmer(
        "-a MY_ADAPTER=GCCGAACTTCTTAGACTGCCTTAAGGACGT",
        "illumina.fastq",
        "illumina.fastq.gz",
    )


def test_adapter_with_u(run_trimmer):
    run_trimmer(
        "-a GCCGAACUUCUUAGACUGCCUUAAGGACGU", "illumina.fastq", "illumina.fastq.gz"
    )


def test_no_trim(run_trimmer):
    run_trimmer(
        "--no-trim --discard-untrimmed -a CCCTAGTTAAAC", "no-trim.fastq", "small.fastq"
    )


def test_bzip2(run_trimmer):
    run_trimmer("-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq.bz2")


@pytest.mark.skipif(lzma is None, reason="no lzma library")
def test_xz(run_trimmer):
    run_trimmer("-b TTAGACATATCTCCGTCG", "small.fastq", "small.fastq.xz")


def test_qualfile_only(input_data):
    retcode = execute_cli(["-sq", input_data("E3M.qual")])
    assert retcode == ReturnCode.ERROR


def test_no_args():
    assert execute_cli() != ReturnCode.SUCCESS


def test_anchored_no_indels(run_trimmer):
    run_trimmer(
        "-g ^TTAGACATAT --no-indels -e 0.1",
        "anchored_no_indels.fasta",
        "anchored_no_indels.fasta",
    )


def test_anchored_no_indels_wildcard_read(run_trimmer):
    run_trimmer(
        "-g ^TTAGACATAT --match-read-wildcards --no-indels -e 0.1",
        "anchored_no_indels_wildcard.fasta",
        "anchored_no_indels.fasta",
    )


def test_anchored_no_indels_wildcard_adapt(run_trimmer):
    run_trimmer(
        "-g ^TTAGACANAT --no-indels -e 0.1",
        "anchored_no_indels.fasta",
        "anchored_no_indels.fasta",
    )


def test_unconditional_cut_front(run_trimmer):
    run_trimmer("-u 5", "unconditional-front.fastq", "small.fastq")


def test_unconditional_cut_back(run_trimmer):
    run_trimmer("-u -5", "unconditional-back.fastq", "small.fastq")


def test_unconditional_cut_both(run_trimmer):
    run_trimmer("-u -5 -u 5", "unconditional-both.fastq", "small.fastq")


def test_untrimmed_output(run_trimmer):
    run_trimmer(
        ["-a", "TTAGACATATCTCCGTCG", "--untrimmed-output", "small.untrimmed.fastq"],
        "small.trimmed.fastq",
        "small.fastq",
        expected_other=["small.untrimmed.fastq"],
    )


def test_adapter_file(run_trimmer, input_data):
    run_trimmer(
        "-a " + input_data.url("adapter.fasta"), "illumina.fastq", "illumina.fastq.gz"
    )


def test_adapter_file_5p_anchored(run_trimmer, input_data):
    run_trimmer(
        "-N -g " + input_data.url("prefix-adapter.fasta"),
        "anchored.fasta",
        "anchored.fasta",
    )


def test_adapter_file_3p_anchored(run_trimmer, input_data):
    run_trimmer(
        "-N -a " + input_data.url("suffix-adapter.fasta"),
        "anchored-back.fasta",
        "anchored-back.fasta",
    )


def test_adapter_file_5p_anchored_no_indels(run_trimmer, input_data):
    run_trimmer(
        "-N --no-indels -g " + input_data.url("prefix-adapter.fasta"),
        "anchored.fasta",
        "anchored.fasta",
    )


def test_adapter_file_3p_anchored_no_indels(run_trimmer, input_data):
    run_trimmer(
        "-N --no-indels -a " + input_data.url("suffix-adapter.fasta"),
        "anchored-back.fasta",
        "anchored-back.fasta",
    )


def test_demultiplex(run_trimmer):
    run_trimmer(
        ["-a", "first=AATTTCAGGAATT", "-a", "second=GTTCTCTAGTTCT"],
        "twoadapters.{name}.fasta",
        "twoadapters.fasta",
        expected_multi=("first", "second", "unknown"),
    )


def test_max_n(run_trimmer):
    run_trimmer("--max-n 0", "maxn0.fasta", "maxn.fasta")
    run_trimmer("--max-n 1", "maxn1.fasta", "maxn.fasta")
    run_trimmer("--max-n 2", "maxn2.fasta", "maxn.fasta")
    run_trimmer("--max-n 0.2", "maxn0.2.fasta", "maxn.fasta")
    run_trimmer("--max-n 0.4", "maxn0.4.fasta", "maxn.fasta")


def test_quiet_is_quiet(capsys, input_data, tmp_path):
    LOGGING_CONFIG.reset()
    tmp_out = tmp_path / "test.fq"
    execute_cli(
        [
            "-o",
            str(tmp_out),
            "--quiet",
            "-a",
            "XXXX",
            "-se",
            str(input_data("illumina.fastq.gz")),
        ]
    )
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def test_twocolor(run_trimmer):
    run_trimmer("--twocolor-trim 22", "nextseq.fastq", "nextseq.fastq")


def test_linked(run_trimmer):
    run_trimmer("-a AAAAAAAAAA...TTTTTTTTTT", "linked.fasta", "linked.fasta")


def test_fasta(run_trimmer):
    run_trimmer("-a TTAGACATATCTCCGTCG", "small.fasta", "small.fastq")


def test_custom_bisulfite_1(run_trimmer):
    run_trimmer(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,1", "small.fastq", "small.fastq"
    )


def test_custom_bisulfite_2(run_trimmer):
    run_trimmer(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 15,15,1,1",
        "small_mincut1.fastq",
        "small.fastq",
    )


def test_custom_bisulfite_3(run_trimmer):
    run_trimmer(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,1,0",
        "small_mincut2.fastq",
        "small.fastq",
    )


def test_custom_bisulfite_4(run_trimmer):
    run_trimmer(
        "-b TTAGACATATCTCCGTCG -q 0,0 --bisulfite 2,2,0,0",
        "small_mincut3.fastq",
        "small.fastq",
    )


@pytest.mark.skipif(
    (
        no_internet("https://ncbi.nlm.nih.gov")
        or no_import("ngstream")
        or no_import("ngs")
    ),
    reason="No internet connection or ngstream not importable",
)
def test_sra(run_trimmer):
    run_trimmer(
        "-b CTGGAGTTCAGACGTGTGCTCT --max-reads 100",
        "SRR2040662_trimmed.fq",
        sra_accn="SRR2040662",
    )



from abc import ABCMeta, abstractmethod
from enum import Enum
from functools import lru_cache
import re
from typing import List, Optional, Pattern, Union

from atropos.io.sequence import Sequence
from atropos.utils import classproperty
from atropos.utils.collections import (
    BaseMergeableDict,
    CountingDict,
    Mergeable,
    NestedDict,
    Summarizable,
    ordered_dict,
)
from atropos.utils.statistics import Histogram
from atropos.utils.ngs import qual2int


DEFAULT_TILE_KEY_REGEXP = r"^(?:[^\:]+\:){4}([^\:]+)"
"""Regexp for the default Illumina read name format."""


class MetricsMode(Enum):
    PRE = "pre"
    POST = "post"


class PositionDicts(Mergeable, Summarizable, metaclass=ABCMeta):
    """
    A sequence of dicts, one for each position in a sequence.
    """

    @classproperty
    def _create_dict(cls) -> BaseMergeableDict:
        pass

    def __init__(self, is_qualities: bool = False, quality_base: int = 33):
        """
        Args:
            is_qualities: Whether values are base qualities.
            quality_base: Base for quality values.
        """
        self.dicts: List[BaseMergeableDict] = []
        self.is_qualities = is_qualities
        self.quality_base = quality_base

    def __getitem__(self, idx: int) -> dict:
        if idx >= len(self.dicts):
            self.extend(idx + 1)
        return self.dicts[idx]

    def extend(self, size: int) -> None:
        """
        Extends the number of bases to `size`.
        """
        diff = size - len(self.dicts)
        if diff > 0:
            for _ in range(diff):
                self.dicts.append(self._create_dict())

    def merge(self, other: "PositionDicts") -> None:
        if not isinstance(other, PositionDicts):
            raise ValueError(f"Cannot merge object of type {type(other)}")

        other_len = len(other.dicts)
        min_len = min(len(self.dicts), other_len)

        for i in range(min_len):
            self.dicts[i].merge(other.dicts[i])

        if other_len > min_len:
            self.dicts.extend(other.dicts[min_len:other_len])

    @abstractmethod
    def summarize(self) -> dict:
        pass


class BaseCountingDicts(PositionDicts):
    """
    A PositionDicts in which the items are CountingDicts that count the number of
    items associated with each nucleotide base.
    """

    @classproperty
    def _create_dict(cls) -> BaseMergeableDict:
        return CountingDict()

    def summarize(self) -> dict:
        """
        Flattens into a table with N rows (where N is the size of the sequence) and the
        columns are counts by nucleotide.

        Returns:
            A tuple of (columns, [rows]), where each row is
            (position, (base_counts...))
        """
        keys = set()

        for dict_item in self.dicts:
            keys.update(dict_item.keys())

        if self.is_qualities:
            keys = tuple(sorted(keys))
            columns = tuple(qual2int(k, self.quality_base) for k in keys)
        else:
            acgt = tuple("ACGT")
            n_val = tuple("N")
            columns = keys = acgt + tuple(keys - set(acgt + n_val)) + n_val

        return dict(
            columns=columns,
            rows=ordered_dict(
                (idx, tuple(dict_item.get(key, 0) for key in keys))
                for idx, dict_item in enumerate(self.dicts, 1)
            ),
        )


class BaseNestedDicts(PositionDicts):
    """
    A PositionDicts in which items are NestedDicts.
    """

    @classproperty
    def _dict_class(cls) -> BaseMergeableDict:
        return NestedDict()

    def summarize(self) -> dict:
        """
        Flattens into a table of N*K rows, where N is the sequence size and K is the
        union of keys in the nested dicts, and the columns are counts by nucleotide.
        """
        keys1 = set()
        keys2 = set()

        for dict1 in self.dicts:
            keys1.update(dict1.keys())
            for dict2 in dict1.values():
                keys2.update(dict2.keys())

        keys1 = tuple(sorted(keys1))
        keys2 = tuple(sorted(keys2))

        if self.is_qualities:
            columns = tuple(qual2int(k, self.quality_base) for k in keys2)
        else:
            columns = keys2

        return dict(
            columns=columns,
            columns2=keys1,
            rows=ordered_dict(
                (
                    idx,
                    ordered_dict(
                        (key1, tuple(dict_item[key1].get(key2, 0) for key2 in keys2))
                        for key1 in keys1
                    ),
                )
                for idx, dict_item in enumerate(self.dicts, 1)
            ),
        )


class ReadMetricsCollector(Summarizable):
    """
    Accumulates metrics on sequencing reads.

    Todo: positional k-mer profiles
    """

    def __init__(
        self,
        qualities=None,
        quality_base: int = 33,
        tiles: Union[bool, str, Pattern] = None,
    ):
        """
        Args:
            qualities: Whether to collect base quality statistics.
            tiles: Whether to collect tile-level statistics. If True, the default
                regular expression is used, otherwise must be a regular expression
                string or compiled re for extracting the tile ID from the read name.
                Only applies to Illumina sequences.
        """
        # max read length
        self.max_read_len = 0
        # read count
        self.count = 0
        # read length distribution
        self.sequence_lengths = Histogram()
        # per-sequence GC percentage
        self.sequence_gc = Histogram()
        # per-position base composition
        self.bases = BaseCountingDicts()
        # whether to collect base quality metrics
        self.qualities = qualities
        self.quality_base = quality_base
        self.tile_key_regexp = None
        self.sequence_qualities = None
        self.base_qualities = None
        self.tile_base_qualities = None

        if qualities:
            tile_key_regexp = DEFAULT_TILE_KEY_REGEXP if tiles is True else tiles
            if isinstance(tile_key_regexp, str):
                tile_key_regexp = re.compile(tile_key_regexp)
            self.tile_key_regexp = tile_key_regexp
            self._init_qualities()

    def _init_qualities(self) -> None:
        # per-sequence mean qualities
        self.sequence_qualities = Histogram()
        # per-position quality composition
        self.base_qualities = BaseCountingDicts(
            is_qualities=True, quality_base=self.quality_base
        )
        if self.tile_key_regexp:
            self.tile_base_qualities = BaseNestedDicts(
                is_qualities=True, quality_base=self.quality_base
            )
            self.tile_sequence_qualities = NestedDict()

    @property
    @lru_cache(maxsize=None)
    def gc_pct(self) -> float:
        return sum(base["C"] + base["G"] for base in self.bases) / self.total_bases

    @property
    @lru_cache(maxsize=None)
    def total_bases(self) -> int:
        return sum(
            length * count for base in self.bases for length, count in base.items()
        )

    @property
    def track_tiles(self) -> bool:
        """
        Whether tile statistics are being accumulated.
        """
        return self.qualities and self.tile_key_regexp is not None

    def collect_record(self, record: Sequence):
        """
        Collects metrics on a single sequence record.
        """
        if self.qualities is None and record.qualities:
            self.qualities = True
            self._init_qualities()

        seq = record.sequence
        seqlen = len(seq)

        self.count += 1
        self.sequence_lengths[seqlen] += 1

        if seqlen > 0:
            gc_pct = round((seq.count("C") + seq.count("G")) * 100 / seqlen)
            self.sequence_gc[gc_pct] += 1

            if seqlen > self.max_read_len:
                self._extend_bases(seqlen)

            tile = None
            if self.track_tiles:
                tile_match = self.tile_key_regexp.match(record.name)
                if tile_match:
                    tile = tile_match.group(1)
                else:
                    raise ValueError(
                        f"{self.tile_key_regexp} did not match {record.name}"
                    )

            if self.qualities:
                quals = record.qualities

                # mean read quality
                # NOTE: we use round here, as opposed to FastQC which uses
                # floor, resulting in slightly different quality profiles
                meanqual = round(
                    sum(ord(q) - self.quality_base for q in quals) / seqlen
                )
                self.sequence_qualities[meanqual] += 1

                if self.track_tiles:
                    self.tile_sequence_qualities[tile][meanqual] += 1

                for i, (base, qual) in enumerate(zip(seq, quals)):
                    self.add_base(i, base, qual, tile)
            else:
                for i, base in enumerate(seq):
                    self.add_base(i, base, tile=tile)

    def add_base(self, i, base, qual=None, tile=None):
        """
        Adds per-base information.

        Args:
            i: Position
            base: Nucleotide
            qual: Quality
            tile: Tile ID
        """
        self.bases[i][base] += 1
        if qual:
            self.base_qualities[i][qual] += 1
            if tile:
                self.tile_base_qualities[i][tile][qual] += 1

    def _extend_bases(self, new_size):
        self.bases.extend(new_size)
        if self.qualities:
            self.base_qualities.extend(new_size)
            if self.track_tiles:
                self.tile_base_qualities.extend(new_size)

    def summarize(self):
        """
        Returns a summary dict.
        """
        summary = dict(
            counts=self.count,
            lengths=self.sequence_lengths.summarize(),
            gc=self.sequence_gc.summarize(),
            bases=self.bases,
        )
        if self.sequence_qualities:
            summary["qualities"] = self.sequence_qualities
        if self.base_qualities:
            summary["base_qualities"] = self.base_qualities
        if self.track_tiles:
            summary["tile_base_qualities"] = self.tile_base_qualities
            summary["tile_sequence_qualities"] = self.tile_sequence_qualities
        return summary


class ReadMetrics(Summarizable, metaclass=ABCMeta):
    @abstractmethod
    def collect(self, read1: Sequence, read2: Optional[Sequence] = None):
        """
        Collects statistics on a pair of reads.
        """


class SingleEndReadMetrics(ReadMetrics):
    """
    ReadStatistics for single-end data.
    """

    def __init__(self, **kwargs):
        self.read1 = ReadMetricsCollector(**kwargs)

    def collect(self, read1: Sequence, read2: Optional[Sequence] = None):
        self.read1.collect_record(read1)

    def summarize(self):
        return dict(read1=super().summarize())


class PairedEndReadMetrics(ReadMetrics):
    """
    ReadStatistics for paired-end data.
    """

    def __init__(self, **kwargs):
        self.read1 = ReadMetricsCollector(**kwargs)
        self.read2 = ReadMetricsCollector(**kwargs)

    def collect(self, read1: Sequence, read2: Optional[Sequence] = None):
        """Collect statistics on a pair of reads.
        """
        self.read1.collect_record(read1)
        self.read2.collect_record(read2)

    def summarize(self):
        """Returns a summary dict.
        """
        return dict(read1=self.read1.summarize(), read2=self.read2.summarize())



from abc import ABCMeta, abstractmethod
from collections.abc import Sequence as SequenceCollection
import copy
from functools import lru_cache
import platform
import sys
from typing import Any, Dict, Iterator, Optional, Sequence as SequenceType, Tuple, cast

from atropos import __version__
from atropos.adapters import AdapterCache
from atropos.errors import AtroposError
from atropos.io.readers import Sequence, open_reader
from atropos.utils import ReturnCode, classproperty
from atropos.utils.argparse import Namespace
from atropos.utils.collections import Const, MergingDict, Summarizable, Timing
from atropos.utils.progress import ProgressBarType, create_progress_reader


EXCEPTION_KEY = "exception"

Batch = Tuple[dict, SequenceType[Sequence]]


class Summary(MergingDict):
    """
    Container for summary information.
    """

    @property
    def has_exception(self) -> bool:
        return EXCEPTION_KEY in self

    def finish(self) -> None:
        self._post_process_dict(self)

    def _post_process_dict(self, dict_val: Optional[Dict[str, Any]]) -> None:
        """
        Replaces `Summarizable` members with their summaries, and computes some
        aggregate values.

        Args:
            dict_val:
        """
        if dict_val is None:
            return

        for key, value in tuple(dict_val.items()):
            if value is None:
                continue

            if isinstance(value, Summarizable):
                dict_val[key] = value = value.summarize()

            if isinstance(value, dict):
                self._post_process_dict(value)
            elif (
                isinstance(value, SequenceCollection)
                and len(value) > 0
                and all(val is None or isinstance(val, dict) for val in value)
            ):
                for val in value:
                    self._post_process_dict(val)
            else:
                if isinstance(value, Const):
                    dict_val[key] = value = value.value
                self._post_process_other(dict_val, key, value)

    def _post_process_other(self, parent: dict, key: str, value: Any) -> None:
        """
        Override this method to perform additional post-processing on non-dict values.

        Args:
            parent: The parent dict
            key: The key
            value: The current value
        """


class Command(Iterator[Batch], metaclass=ABCMeta):
    """
    An Atropos command.
    """

    @abstractmethod
    def get_option(self, name: str, default: Optional[Any] = None) -> Any:
        """
        Gets the value of the specified option.

        Args:
            name: The option name.
            default: The default value to return if option `name` is not set.

        Returns:
            The option value.
        """

    @abstractmethod
    def iterator(self) -> Iterator[Batch]:
        """
        Returns an iterator over input batches.
        """

    @abstractmethod
    def merge_summary(self, summary: dict) -> None:
        """
        Merge a summary dict into the command's current summary.

        Args:
            summary: The summary to merge
        """


class BaseCommand(Command, metaclass=ABCMeta):
    """
    Base class for Atropos commands.
    """

    @classproperty
    @abstractmethod
    def name(cls) -> str:
        """
        The command name.
        """

    @classmethod
    def _create_summary(cls) -> dict:
        return Summary()

    def __init__(self, options: Namespace):
        """
        Args:
            options: Command-line options.
        """
        self.options = options
        self.summary: dict = self._create_summary()
        self.timing = Timing()
        self.return_code: Optional[ReturnCode] = None
        self.size = options.batch_size or 1000
        self.batches = 0
        self.done = False
        self._empty_batch = [None] * self.size
        self._progress_options = None

        open_args = dict(
            file_format=options.input_format,
            quality_base=options.quality_base,
            colorspace=options.colorspace,
            input_read=options.input_read,
            alphabet=options.alphabet,
        )

        if options.ngstream_reader:
            open_args["ngstream_reader"] = options.ngstream_reader
            options.ngstream_reader = None
        else:
            interleaved = bool(options.interleaved_input)
            input1 = options.interleaved_input if interleaved else options.input1
            input2 = qualfile = None

            if options.paired and not interleaved:
                input2 = options.input2
            else:
                qualfile = options.input2

            open_args.update(
                file1=input1,
                file2=input2,
                qualfile=qualfile,
                interleaved=interleaved,
            )

        if options.umi_input:
            open_args.update(umi_file=options.umi_input)

        self.reader = reader = open_reader(**open_args)

        # Wrap reader in subsampler
        if options.subsample:
            import random

            if options.subsample_seed:
                random.seed(options.subsample_seed)

            def subsample(_reader, frac):
                """Generator that yields a random subsample of records.

                Args:
                    _reader: The reader from which to sample.
                    frac: The fraction of records to yield.
                """
                for reads in _reader:
                    if random.random() < frac:
                        yield reads

            reader = subsample(reader, options.subsample)

        self.iterable = enumerate(reader, 1)

        if options.progress != ProgressBarType.NONE:
            max_reads = self.get_option("max_reads")

            if not max_reads:
                max_reads = self.reader.estimate_num_records()
                if max_reads and options.subsample:
                    max_reads = int(max_reads * options.subsample)

            self._progress_options = (
                options.progress,
                self.size,
                max_reads,
                options.counter_magnitude,
            )

        self._init_summary()

    @lru_cache(maxsize=None)
    def get_option(self, name: str, default: Optional[Any] = None) -> Any:
        if hasattr(self.reader, name):
            return getattr(self.reader, name)
        elif hasattr(self.options, name):
            return getattr(self.options, name)
        else:
            return default

    def iterator(self) -> Iterator[Batch]:
        """
        Returns an iterator (an object with the __iter__ method) over input batches.
        BaseCommand is itself an iterable, and will be wrapped with a progress bar if
        `_progress_options` is True.
        """
        if self._progress_options:
            # Wrap iterator in progress bar
            return create_progress_reader(self, *self._progress_options)
        else:
            return self

    def merge_summary(self, summary: dict) -> None:
        if isinstance(self.summary, MergingDict):
            cast(MergingDict, self.summary).merge(summary)
        else:
            self.summary.update(summary)

    def __iter__(self):
        return self

    def __next__(self) -> Batch:
        if self.done:
            raise StopIteration()

        try:
            read_index, record = next(self.iterable)
        except StopIteration:
            self.finish()
            raise

        batch = copy.copy(self._empty_batch)
        batch[0] = record
        batch_index = 1
        max_reads = self.get_option("max_reads")
        if max_reads:
            max_size = min(self.size, max_reads - read_index + 1)
        else:
            max_size = self.size

        while batch_index < max_size:
            try:
                read_index, record = next(self.iterable)
                batch[batch_index] = record
                batch_index += 1
            except StopIteration:
                self.finish()
                break
            except:
                self.finish()
                raise

        if max_reads and read_index >= max_reads:
            self.finish()

        self.batches += 1

        batch_meta = dict(
            index=self.batches,
            # TODO: When multi-file input is supported, 'source' will need to
            # be the index of the current file/pair from which records are
            # being read.
            source=0,
            size=batch_index,
        )

        if batch_index == self.size:
            return batch_meta, batch
        else:
            return batch_meta, batch[0:batch_index]

    def _init_summary(self) -> None:
        """
        Initializes the summary dict with general information.
        """
        self.summary["program"] = "Atropos"
        self.summary["version"] = __version__
        self.summary["python"] = platform.python_version()
        self.summary["command"] = self.name
        self.summary["options"] = self.options.__dict__.copy()
        self.summary["timing"] = self.timing
        self.summary["sample_id"] = self.options.sample_id
        self.summary["input"] = self.reader.summarize()
        self.summary["input"].update(
            batch_size=self.size,
            max_reads=self.get_option("max_reads"),
            batches=self.batches
        )

    def run(self) -> Tuple[ReturnCode, dict]:
        """
        Runs the command, wrapping it in a Timing, catching any exceptions,
        and finally closing the reader.

        Returns:
            The tuple (retcode, summary).
        """
        with self.timing:
            try:
                self.return_code = self()
            except Exception as err:  # pylint: disable=broad-except
                self.summary["exception"] = dict(
                    message=str(err), details=sys.exc_info()
                )
                self.return_code = 1
            finally:
                self.finish()
        return self.return_code, self.summary

    @abstractmethod
    def __call__(self) -> ReturnCode:
        """
        Executes the command.

        Returns:
            The return code.
        """

    def finish(self) -> None:
        """
        Finishes the command.
        """
        # Close the underlying reader.
        if not self.done:
            self.done = True
            self.reader.close()
        if isinstance(self.summary, Summary):
            cast(Summary, self.summary).finish()

    def _load_known_adapters(self) -> AdapterCache:
        """
        Load known adapters based on setting in command-line options.
        """
        cache_file = None
        if self.options.cache_adapters:
            cache_file = self.options.adapter_cache_file
        adapter_cache = AdapterCache(cache_file)
        if adapter_cache.empty and self.options.default_adapters:
            adapter_cache.load_default()
        if self.options.known_adapter:
            for known in self.options.known_adapter:
                name, seq = known.split("=")
                adapter_cache.add(name, seq)
        if self.options.known_adapters_file:
            for known_file in self.options.known_adapters_file:
                adapter_cache.load_from_url(known_file)
        if self.options.cache_adapters:
            adapter_cache.save()
        return adapter_cache


class Pipeline(metaclass=ABCMeta):
    """
    Base class for analysis pipelines.
    """

    def __init__(self):
        self.record_counts = {}
        self.bp_counts = {}

    def __call__(
        self, command_runner: BaseCommand, raise_on_error: bool = False, **kwargs
    ) -> None:
        """
        Executes the pipeline.

        Args:
            command_runner:
            raise_on_error:
            **kwargs:

        Raises:

        """
        self.start(**kwargs)
        try:
            for batch in command_runner.iterator():
                self.process_batch(batch)
        except Exception as err:
            if raise_on_error:
                raise
            else:
                command_runner.summary["exception"] = dict(
                    message=str(err), details=sys.exc_info()
                )
        finally:
            self.finish(command_runner.summary, **kwargs)

    def start(self, **kwargs) -> None:
        """
        Starts the pipeline.
        """

    def process_batch(self, batch: Batch) -> None:
        """
        Runs the pipeline on a batch of records.

        Args:
            batch: A batch of reads. A batch has the format
            ({batch_metadata}, [records]).
        """
        batch_meta, records = batch
        context = batch_meta.copy()
        if context["source"] not in self.record_counts:
            self.record_counts[context["source"]] = 0
        self.record_counts[context["source"]] += context["size"]
        if not context["source"] in self.bp_counts:
            self.bp_counts[context["source"]] = [0, 0]
        context["bp"] = self.bp_counts[context["source"]]
        self.add_to_context(context)
        self.handle_records(context, records)

    def add_to_context(self, context: dict) -> None:
        """
        Adds items to the batch context.
        """

    def handle_records(self, context: dict, records: SequenceType[Sequence]) -> None:
        """
        Handles a sequence of records.

        Args:
            context: The pipeline context (dict).
            records: The sequence of records.

        Raises:
            AtroposError
        """
        for idx, record in enumerate(records):
            try:
                self.handle_record(context, record)
            except Exception as err:
                raise AtroposError(
                    f"An error occurred at record {idx} of batch {context['index']}"
                ) from err

    @abstractmethod
    def handle_record(self, context: dict, record: Sequence) -> None:
        """
        Handles a single record.

        Args:
            context: The pipeline context (dict).
            record: The record.
        """

    @abstractmethod
    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ) -> None:
        """
        Handles a read or read-pair.

        Args:
            context: The pipeline context (dict).
            read1: The first read in the pair.
            read2: The second read in the pair; will be `None` for single-end data.
        """

    def finish(self, summary: dict, **kwargs) -> None:
        """
        Finishes the pipeline, including adding information to the summary.

        Args:
            summary: Summary dict to update.
        """
        total_bp_counts = tuple(sum(b) for b in zip(*self.bp_counts.values()))
        summary.update(
            record_counts=self.record_counts,
            total_record_count=sum(self.record_counts.values()),
            bp_counts=self.bp_counts,
            total_bp_counts=total_bp_counts,
            sum_total_bp_count=sum(total_bp_counts),
        )


class SingleEndPipelineMixin:
    """
    Mixin for pipelines that implements `handle_record` for single-end data.
    """

    def handle_record(self, context: dict, record: Sequence) -> None:
        context["bp"][0] += len(record)
        cast(Pipeline, self).handle_reads(context, record)


class PairedEndPipelineMixin:
    """
    Mixin for pipelines that implements `handle_record` for paired-end data.
    """

    def handle_record(self, context: dict, record: Sequence) -> None:
        read1, read2 = record
        bps = context["bp"]
        bps[0] += len(read1.sequence)
        bps[1] += len(read2.sequence)
        cast(Pipeline, self).handle_reads(context, read1, read2)



from enum import Enum
from typing import Iterable, Iterator, Optional, TypeVar

import pokrok


T = TypeVar("T")


class ProgressBarType(Enum):
    BAR = "bar"
    MSG = "msg"
    ANY = "any"
    NONE = "none"


def create_progress_reader(
    itr: Iterable[T],
    progress_type: ProgressBarType = ProgressBarType.BAR,
    batch_size: int = 1,
    max_items: Optional[int] = None,
    **kwargs
) -> Iterator[T]:
    """
    Wraps an iterable in a progress bar of the specified type.

    Args:
        itr: The iterable to wrap.
        progress_type: msg = a custom progress bar that reports via log
            messages; bar = use a ProgressBar (from the progressbar library)
            or tqdm.
        max_items: Max number of items, if known in advance.
        batch_size: The number of records in each iterable item (iterable is
            typically a BatchReader).
        kwargs: Additional arguments to pass to the progress bar constructor.

    Returns:
        A wrapped iterable. If `progress_type == 'bar'` and neither of the
        supported libraries are available, a warning is logged and the unwrapped
        reader is returned.
    """
    if progress_type == ProgressBarType.NONE:
        return iter(itr)

    if progress_type == ProgressBarType.MSG:
        pokrok.set_plugins(["logging"])
    elif progress_type == ProgressBarType.BAR:
        pokrok.set_plugins(["tqdm", "progressbar2"])

    if max_items:
        widgets = [
            pokrok.Widget.COUNTER,
            pokrok.Widget.PERCENT,
            pokrok.Widget.ELAPSED,
            pokrok.Widget.BAR,
            pokrok.Widget.ETA,
        ]
    else:
        widgets = [pokrok.Widget.COUNTER, pokrok.Widget.ELAPSED, pokrok.Widget.SPINNER]

    return pokrok.progress_iter(
        itr,
        desc="Processed",
        size=max_items,
        unit="records",
        multiplier=batch_size,
        style=pokrok.Style(widgets),
        **kwargs,
    )



from abc import ABCMeta, abstractmethod
from collections import defaultdict
import csv
import re
from pathlib import Path
from typing import Optional, Tuple, Union

from xphyle import open_

from atropos.errors import AtroposError
from atropos.commands import (
    BaseCommand,
    Pipeline,
    SingleEndPipelineMixin,
    PairedEndPipelineMixin,
)
from atropos.io.sequence import Sequence
from atropos.utils import ReturnCode, classproperty, run_interruptible
from atropos.utils.ngs import qual2prob


class ErrorCommand(BaseCommand):
    @classproperty
    def name(cls) -> str:
        return "error"

    def __call__(self) -> ReturnCode:
        if not self.get_option("delivers_qualities"):
            raise ValueError("Cannot estimate error rate without base qualities")

        algorithm = self.get_option("algorithm")
        if algorithm == "quality":
            estimator_class = BaseQualityErrorEstimator
        elif algorithm == "shadow":
            estimator_class = ShadowRegressionErrorEstimator
        else:
            raise ValueError(f"Invalid algorithm: {algorithm}")

        estimator_args = dict(max_read_len=self.get_option("max_bases"))

        if self.get_option("paired"):
            estimator = PairedErrorEstimator(
                estimator_class=estimator_class, **estimator_args
            )
        else:
            estimator = estimator_class(**estimator_args)

        self.summary["errorrate"] = estimator_args
        # currently only single-threaded operation is supproted
        self.summary.update(mode="serial", threads=1)

        return run_interruptible(estimator, self, raise_on_error=True)


class ErrorEstimator(SingleEndPipelineMixin, Pipeline, metaclass=ABCMeta):
    """
    Base class for error estimators.
    """

    def __init__(self, max_read_len: int):
        super().__init__()
        self.total_len = 0
        self.max_read_len = max_read_len

    @abstractmethod
    def estimate(self) -> Tuple[float, Optional[dict]]:
        """
        Estimate the error rate.
        """

    def finish(self, summary: dict, **kwargs):
        super().finish(summary)
        estimate, details = self.estimate()
        summary["errorrate"].update(
            estimate=(estimate,), total_len=(self.total_len,), details=(details,)
        )


class BaseQualityErrorEstimator(ErrorEstimator):
    """
    Simple error estimation using base qualities. It is well-known that base
    qualities significantly overestimate true error rates, so take the estimate with
    a grain of salt.
    """

    def __init__(self, max_read_len=None):
        super().__init__(max_read_len)
        self.total_qual = 0.0

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        quals = read1.qualities
        readlen = len(quals)
        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            quals = quals[:readlen]
        self.total_qual += sum(qual2prob(qchar) for qchar in quals)
        self.total_len += readlen

    def estimate(self) -> Tuple[float, Optional[dict]]:
        return self.total_qual / self.total_len, None


# Error estimation using shadow counts
# DEPRECATED
# TODO: Replace with Zhu et al. cubic splines method or something faster
# https://bmcbioinformatics.biomedcentral.com/articles/10.1186/s12859-016-1052-3

FILTER_RE = re.compile("A+|C+|G+|T+|.*N.*")
SHADOW_REGRESSION_R_SCRIPT = """
library(ShadowRegression)
errorRates = getErrorRates("{reads}", type="{method}")
write.table(errorRates$perReadER, "{per_read}", sep="\t", quote=F, col.names=F,
            row.names=T)
write.table(errorRates$cycleER, "{per_cycle}", sep="\t", quote=F, col.names=F,
            row.names=T)
"""


class ShadowRegressionErrorEstimator(ErrorEstimator):
    """
    Re-implementation of the shadow regression method described in:
    Wang et al., "Estimation of sequencing error rates in short reads",
        BMC Bioinformatics 2012 13:185, DOI: 10.1186/1471-2105-13-185
    """

    def __init__(
        self,
        method: str = "sub",
        max_read_len: Optional[int] = None,
        rscript_exe: Union[str, Path] = "Rscript"
    ):
        """
        Args:
            method: The differences that are considered in the error rate
                calculation; sub = substitutions, indel = insertions and
                deletions; all = both substitutions and indels.
            max_read_len: The maximum number of bases (starting from the 5' end)
                to consider from each read.
            rscript_exe:
        """
        super().__init__(max_read_len)
        self.seqs = defaultdict(lambda: 0)
        self.method = method
        self.rscript_exe = rscript_exe

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        seq = read1.sequence
        readlen = len(seq)

        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            seq = seq[:readlen]

        if FILTER_RE.fullmatch(seq):
            return

        self.seqs[seq] += 1
        self.total_len += readlen

    def estimate(self) -> Tuple[float, Optional[dict]]:
        # This is a temporary solution that requires R and the
        # ShadowRegression package. Eventually, this will be
        # replaced with a pure-python implementation.
        import subprocess
        import tempfile

        tempfiles = (Path(tempfile.mkstemp()[1]) for _ in range(4))
        read_counts, per_read, per_cycle, script_file = tempfiles

        try:
            # write counts to a file
            with open_(read_counts, "wt") as out:
                self._write_read_counts(out)

            # execute R script
            script = SHADOW_REGRESSION_R_SCRIPT.format(
                reads=read_counts,
                method=self.method,
                per_read=per_read,
                per_cycle=per_cycle,
            )

            with open_(script_file, "wt") as out:
                out.write(script)

            proc = subprocess.Popen(
                [str(self.rscript_exe), "--vanilla", script_file],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            with proc:
                stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    raise AtroposError(
                        f"R script failed: rc={proc.returncode}; stdout={stdout}; "
                        f"stderr={stderr}"
                    )

            # read the results
            with open_(per_read, "rt") as i:
                reader = csv.reader(i, delimiter="\t")
                per_read_error = dict(reader)
                if len(per_read_error) != 4:
                    raise AtroposError("Invalid output from R script")

            with open_(per_cycle, "rt") as i:
                reader = csv.reader(i, delimiter="\t")
                per_cycle_error = list(row[0:3] for row in reader)
                if not per_cycle_error:
                    raise AtroposError("Invalid output from R script")

            return (
                per_read_error["error rate"],
                dict(per_read=per_read_error, per_cycle=per_cycle_error),
            )

        finally:
            for path in tempfiles:
                path.unlink()

    def _write_read_counts(self, fileobj):
        writer = csv.writer(fileobj, delimiter=" ")
        writer.writerows(sorted(self.seqs.items(), reverse=True, key=lambda i: i[1]))


class PairedErrorEstimator(PairedEndPipelineMixin, Pipeline):
    """
    Estimator for a pair of input files.
    """

    def __init__(self, estimator_class=BaseQualityErrorEstimator, **kwargs):
        super().__init__()
        self.estimator1 = estimator_class(**kwargs)
        self.estimator2 = estimator_class(**kwargs)

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        self.estimator1.handle_reads(context, read1)
        self.estimator2.handle_reads(context, read2)

    def finish(self, summary, **kwargs):
        """
        Estimates error rates.

        Returns:
            Tuple (read1_estimate, read2_estimate).
        """
        super().finish(summary)
        estimate1, details1 = self.estimator1.estimate()
        estimate2, details2 = self.estimator2.estimate()
        summary["errorrate"].update(
            estimate=(estimate1, estimate2),
            total_len=(self.estimator1.total_len, self.estimator2.total_len),
            details=(details1, details2),
        )



class AtroposError(Exception):
    """
    Base class for Atropos-specific errors.
    """


class FormatError(AtroposError):
    """
    Raised when an input file (FASTA or FASTQ) is malformatted.
    """


class UnknownFileTypeError(AtroposError):
    """
    Raised when file type could not be detected.
    """


class MulticoreError(AtroposError):
    """
    Base error for parallel processes.
    """


class NotInAlphabetError(AtroposError):
    """
    Error that indicates an invalid character in an input sequence.
    """
    def __init__(self, character: str):
        super().__init__()
        self.character = character



from atropos.adapters import AdapterType
from atropos.aligner import (
    Aligner,
    locate,
    compare_prefixes,
    compare_suffixes,
)


class TestAligner:
    def test(self):
        reference = "CTCCAGCTTAGACATATC"
        aligner = Aligner(reference, 0.1, flags=AdapterType.BACK)
        aligner.locate("CC")

    def test_100_percent_error_rate(self):
        reference = "GCTTAGACATATC"
        aligner = Aligner(reference, 1.0, flags=AdapterType.BACK)
        aligner.locate("CAA")


def test_polya():
    s = "AAAAAAAAAAAAAAAAA"
    t = "ACAGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    result = locate(s, t, 0.0, AdapterType.BACK)
    # start_s, stop_s, start_t, stop_t, matches, cost = result
    assert result == (0, len(s), 4, 4 + len(s), len(s), 0)


# Sequences with IUPAC wildcards
# R=A|G, Y=C|T, S=G|C, W=A|T, K=G|T, M=A|C, B=C|G|T, D=A|G|T, H=A|C|T, V=A|C|G,
# N=A|C|G|T, X={}
WILDCARD_SEQUENCES = [
    "CCCATTGATC",  # original sequence without wildcards
    "CCCRTTRATC",  # R=A|G
    "YCCATYGATC",  # Y=C|T
    "CSSATTSATC",  # S=G|C
    "CCCWWWGATC",  # W=A|T
    "CCCATKKATC",  # K=G|T
    "CCMATTGMTC",  # M=A|C
    "BCCATTBABC",  # B=C|G|T
    "BCCATTBABC",  # B
    "CCCDTTDADC",  # D=A|G|T
    "CHCATHGATC",  # H=A|C|T
    "CVCVTTVATC",  # V=A|C|G
    "CCNATNGATC",  # N=A|C|G|T
    "CCCNTTNATC",
    # N
    #   'CCCXTTXATC',  # X
]


def test_compare_prefixes():
    assert compare_prefixes("AAXAA", "AAAAATTTTTTTTT") == (0, 5, 0, 5, 4, 1)
    assert compare_prefixes("AANAA", "AACAATTTTTTTTT", wildcard_ref=True) == (
        0, 5, 0, 5, 5, 0,
    )
    assert compare_prefixes("AANAA", "AACAATTTTTTTTT", wildcard_ref=True) == (
        0, 5, 0, 5, 5, 0,
    )
    assert compare_prefixes("XAAAAA", "AAAAATTTTTTTTT") == (0, 6, 0, 6, 4, 2)

    a = WILDCARD_SEQUENCES[0]

    for s in WILDCARD_SEQUENCES:
        r = s + "GCCAGGGTTGATTCGGCTGATCTGGCCG"
        result = compare_prefixes(a, r, wildcard_query=True)
        assert result == (0, 10, 0, 10, 10, 0), result
        result = compare_prefixes(r, a, wildcard_ref=True)
        assert result == (0, 10, 0, 10, 10, 0)

    for s in WILDCARD_SEQUENCES:
        r = s + "GCCAGGG"
        result = compare_prefixes(s, r)
        assert result == (0, 10, 0, 10, 10, 0)
        result = compare_prefixes(r, s, wildcard_ref=True, wildcard_query=True)
        assert result == (0, 10, 0, 10, 10, 0)

    r = WILDCARD_SEQUENCES[0] + "GCCAGG"
    for wildc_ref in (False, True):
        for wildc_query in (False, True):
            result = compare_prefixes(
                "CCCXTTXATC", r, wildcard_ref=wildc_ref, wildcard_query=wildc_query
            )
            assert result == (0, 10, 0, 10, 8, 2)


def test_compare_suffixes():
    assert compare_suffixes("AAXAA", "TTTTTTTAAAAA") == (0, 5, 7, 12, 4, 1)
    assert compare_suffixes("AANAA", "TTTTTTTAACAA", wildcard_ref=True) == (
        0, 5, 7, 12, 5, 0
    )
    assert compare_suffixes("AANAA", "TTTTTTTAACAA", wildcard_ref=True) == (
        0, 5, 7, 12, 5, 0
    )
    assert compare_suffixes("AAAAAX", "TTTTTTTAAAAA") == (0, 6, 6, 12, 4, 2)


def test_wildcards_in_adapter():
    r = "CATCTGTCC" + WILDCARD_SEQUENCES[0] + "GCCAGGGTTGATTCGGCTGATCTGGCCG"
    for a in WILDCARD_SEQUENCES:
        result = locate(a, r, 0.0, AdapterType.BACK, wildcard_ref=True)
        assert result == (0, 10, 9, 19, 10, 0), result
    a = "CCCXTTXATC"
    result = locate(a, r, 0.0, AdapterType.BACK, wildcard_ref=True)
    assert result is None


def test_wildcards_in_read():
    a = WILDCARD_SEQUENCES[0]
    for s in WILDCARD_SEQUENCES:
        r = "CATCTGTCC" + s + "GCCAGGGTTGATTCGGCTGATCTGGCCG"
        result = locate(a, r, 0.0, AdapterType.BACK, wildcard_query=True)
        if "X" in s:
            assert result is None
        else:
            assert result == (0, 10, 9, 19, 10, 0), result


def test_wildcards_in_both():
    for a in WILDCARD_SEQUENCES:
        for s in WILDCARD_SEQUENCES:
            if "X" in s or "X" in a:
                continue

            r = "CATCTGTCC" + s + "GCCAGGGTTGATTCGGCTGATCTGGCCG"
            result = locate(
                a, r, 0.0, AdapterType.BACK, wildcard_ref=True, wildcard_query=True
            )
            assert result == (0, 10, 9, 19, 10, 0), result


def test_no_match():
    a = locate("CTGATCTGGCCG", "AAAAGGG", 0.1, AdapterType.BACK)
    assert a is None, a


def test_multi_aligner_no_mismatches():
    from atropos.aligner import MultiAligner

    a = MultiAligner(max_error_rate=0, min_overlap=3)
    matches = a.locate("AGAGATCAGATGACAGATC", "GATCA")
    assert len(matches) == 2
    matches.sort(key=lambda x: x[4], reverse=True)
    assert matches[0][0] == 3
    assert matches[0][1] == 8
    assert matches[0][2] == 0
    assert matches[0][3] == 5
    assert matches[0][4] == 5
    assert matches[0][5] == 0
    assert matches[1][0] == 15
    assert matches[1][1] == 19
    assert matches[1][2] == 0
    assert matches[1][3] == 4
    assert matches[1][4] == 4
    assert matches[1][5] == 0


def test_multi_aligner_with_mismatches():
    from atropos.aligner import MultiAligner

    a = MultiAligner(max_error_rate=0.1, min_overlap=10)
    matches = a.locate("GATATCAGATGACAGATCAGAGATCAGAT", "GAGATCAGATGA")
    assert len(matches) == 2
    matches.sort(key=lambda x: x[5])
    assert matches[0][0] == 19
    assert matches[0][1] == 29
    assert matches[0][2] == 0
    assert matches[0][3] == 10
    assert matches[0][4] == 10
    assert matches[0][5] == 0
    assert matches[1][0] == 0
    assert matches[1][1] == 12
    assert matches[1][2] == 0
    assert matches[1][3] == 12
    assert matches[1][4] == 11
    assert matches[1][5] == 1



from enum import IntEnum
from functools import partial
import inspect
from multiprocessing import Process, Value
from queue import Empty, Full, Queue
from threading import Lock
import time
from typing import Any, Callable, Iterable, Optional, Union, cast

from loguru import logger

from atropos.errors import MulticoreError
from atropos.utils import ReturnCode


DEFAULT_RETRY_INTERVAL = 5
DEFAULT_BLOCK_TIMEOUT = 5


class ControlSignal(IntEnum):
    ACTIVE = 0
    """Controlled process should run normally."""
    ERROR = -1
    """Controlled process should exit."""


class Control:
    """
    Shared (long) value for passing control information between main and worker threads.

    Args:
        initial_value: Initial value of the shared control variable.
    """

    def __init__(self, initial_value=ControlSignal.ACTIVE):
        self.control = Value('l', initial_value)

    def check_value(self, value, lock=False):
        """
        Checks that the current control value == `value`.

        Args:
            value: The value to check.
            lock: Whether to lock the shared variable before checking.

        Returns:
            True if the values are equal.
        """
        return self.get_value(lock=lock) == value

    def check_value_positive(self, lock=False):
        """
        Checks that the current control value is positive.

        Args:
            lock: Whether to lock the shared variable before checking.
        """
        return self.get_value(lock=lock) > 0

    def get_value(self, lock=True):
        """
        Returns the current control value.

        Args:
            lock: Whether to lock the shared variable before checking.
        """
        if lock:
            with self.control.get_lock():
                return self.control.value

        else:
            return self.control.value

    def set_value(self, value):
        """
        Sets the control value. The shared variable is always locked.

        Args:
            value: The value to set.
        """
        with self.control.get_lock():
            self.control.value = value


class PendingQueue:
    """
    Queue for items with sequentially increasing priority. An item whose
    priority is below the current level is queued. Pop returns the item with
    the current priority and increments the current priority.
    """

    def __init__(self, max_size=None):
        """
        Args:
            max_size: Maximum queue size; None == infinite.
        """
        self.queue = {}
        self.max_size = max_size
        self.min_priority = None

    def push(self, priority, value):
        """
        Adds an item to the queue with priority.

        Args:
            priority: An integer that determines the placement of `value` in
                the queue. Must be unique.
            value: The value to queue.

        Raises:
            Full if queue is full.
        """
        if self.full:
            raise Full()

        if priority in self.queue:
            raise ValueError(f"Duplicate priority value: {priority}")

        self.queue[priority] = value
        if self.min_priority is None or priority < self.min_priority:
            self.min_priority = priority

    def pop(self):
        """
        Removes and return the item in the queue with lowest priority.

        Raises:
            Empty if the queue is emtpy.
        """
        if self.empty:
            raise Empty()

        value = self.queue.pop(self.min_priority)
        if self.empty:
            self.min_priority = None
        else:
            self.min_priority = min(self.queue.keys())
        return value

    @property
    def full(self):
        """
        Whether the queue is full.
        """
        return self.max_size and len(self.queue) >= self.max_size

    @property
    def empty(self):
        """
        Whether the queue is empty.
        """
        return len(self.queue) == 0


def synchronized(f):
    """Synchronization decorator."""

    f._lock = Lock()

    def wrapper(*args, **kw):
        f._lock.acquire()

        try:
            return f(*args, **kw)
        finally:
            f._lock.release()

    return wrapper


def ensure_processes(
    processes: Iterable[Process],
    message: str = "One or more process exited: {}",
    alive: bool = True
):
    """
    Raises an exception if all processes do not have the expected status (alive or
    dead).
    """
    is_alive = [worker.is_alive() for worker in processes]
    if alive != all(is_alive):
        raise MulticoreError(
            message.format(
                ",".join(str(i) for i, a in enumerate(is_alive) if a != alive)
            )
        )


def wait_on(
    condition: Callable,
    *args,
    wait_message: str = "Waiting {}",
    timeout: Optional[int] = None,
    fail_callback: Optional[Callable] = None,
    wait: Union[bool, int, Callable, None] = None,
    timeout_callback: Optional[Callable] = None,
    retry_interval: int = DEFAULT_RETRY_INTERVAL
) -> Any:
    """
    Waits on a condition to be non-False.

    Args:
        condition: Function that returns either False or a non-False value.
        args: Args with which to call `condition`.
        wait_message: The message to log while `condition` is False.
        timeout: Number of seconds after which the log messages escalate from
            DEBUG to ERROR.
        fail_callback: Function that is called each time `condition` returns
            False.
        wait: Either a boolean or a wait function to execute. If True,
            `time.sleep(RETRY_INTERVAL)` is used as the wait function.
        timeout_callback: Eather a function that is called each time, or an
            Exception class that is raised the first time `condition`
            returns False after waiting for `timeout` seconds.
        retry_interval:

    Returns:

    """
    if wait is True:
        wait_fn = partial(time.sleep, retry_interval)
    elif isinstance(wait, int):
        wait_time = wait
        wait_fn = partial(time.sleep, wait_time)
    elif wait:
        wait_fn = cast(Callable, wait)
    else:
        wait_fn = None

    wait_start = None

    while True:
        result = condition(*args)
        if result is not False:
            return result

        if fail_callback:
            fail_callback()

        now = time.time()

        if not wait_start:
            wait_start = now
        else:
            waiting = now - wait_start
            msg = wait_message.format(f"for {round(waiting, 1)} seconds")

            if timeout is not None and waiting >= timeout:
                logger.error(msg)
                if timeout_callback:
                    if inspect.isclass(timeout_callback):
                        raise timeout_callback()
                    else:
                        timeout_callback()
            else:
                logger.debug(msg)

            if wait_fn:
                wait_fn()


def wait_on_process(
    process: Process,
    wait_timeout: int,
    terminate: bool = False,
    retry_interval: int = DEFAULT_RETRY_INTERVAL
) -> Any:
    """
    Waits on a process to terminate.

    Args:
        process: The process on which to wait.
        wait_timeout: Number of seconds to wait for process to terminate.
        terminate: Whether to force the process to terminate after `timeout`
            seconds.
        retry_interval:
    """
    timeout_callback = process.terminate if terminate else None
    return wait_on(
        lambda: not process.is_alive(),
        wait_message=f"Waiting on {process.name} to terminate {{}}",
        timeout=wait_timeout,
        wait=lambda: process.join(retry_interval),
        timeout_callback=timeout_callback,
    )


def enqueue(
    queue: Queue,
    item: Any,
    wait_message: str = "Waiting to enqueue item {}",
    block_timeout: int = DEFAULT_BLOCK_TIMEOUT,
    **kwargs,
) -> None:
    """
    Enqueues an item, using `wait_on` to wait while `queue` is full.

    Args:
        queue: The queue to which to add the item.
        item: The item to queue.
        wait_message: The message to log while waiting.
        block_timeout: Number of seconds to wait after each `queue.put` attempt.
        kwargs: Additional arguments to `wait_on`.
    """

    def condition(_item):
        """
        Returns True if enqueing was successful.
        """
        try:
            queue.put(_item, block=True, timeout=block_timeout)
            return True
        except Full:
            return False

    wait_on(condition, item, wait_message=wait_message, **kwargs)


def enqueue_all(
    iterable: Iterable,
    queue: Queue,
    wait_timeout: int,
    fail_callback: Callable,
    block_timeout: int = DEFAULT_BLOCK_TIMEOUT
) -> int:
    """
    Enqueues all items in `iterable`, using `wait_on` to wait while `queue` is full.

    Args:
        iterable: Iterable of items to queue.
        queue: The queue to which to add the item.
        wait_timeout: Number of seconds to wait after each `queue.put` attempt.
        fail_callback: Function called (or Exception raised) after timeout.
        block_timeout: Number of seconds to block while attempting to enqueue an item.

    Returns:
        The number of items queued.
    """
    num_items = 0

    def condition(_item):
        """Returns True if enqueing was successful.
        """
        try:
            queue.put(_item, block=True, timeout=block_timeout)
            return True
        except Full:
            return False

    for item in iterable:
        wait_on(
            condition,
            item,
            wait_message="Main process waiting to queue item {}",
            timeout=wait_timeout,
            fail_callback=fail_callback,
        )
        num_items += 1

    return num_items


def dequeue(
    queue: Queue,
    wait_message: str = "Waiting to dequeue item {}",
    block_timeout: int = DEFAULT_BLOCK_TIMEOUT,
    **kwargs,
) -> Any:
    """
    Dequeues an item, using `wait_on` to wait while `queue` is empty.
    """

    def condition():
        """
        Returns an item from the queue, or False if the queue is empty.
        """
        try:
            return queue.get(block=True, timeout=block_timeout)
        except Empty:
            return False

    return wait_on(condition, wait_message=wait_message, **kwargs)


def kill(process: Process, retcode: ReturnCode, wait_timeout: int):
    """
    Kills a process if it fails to terminate on its own.
    """
    if retcode <= 1:
        wait_on_process(process, wait_timeout, terminate=True)
    elif process.is_alive():
        process.terminate()



from enum import IntFlag, Enum
from pathlib import Path
from typing import IO, Set, Union

from atropos.errors import UnknownFileTypeError
from atropos.utils.paths import splitext_compressed


class InputRead(IntFlag):
    SINGLE = 0
    READ1 = 1
    READ2 = 2
    PAIRED = 1 | 2


class SequenceFileType(Enum):
    FASTA = {".fasta", ".fa", ".fna", ".csfasta", ".csfa"}, True
    FASTQ = {".fastq", ".fq"}, True
    SAM = {".sam"}, True
    BAM = {".bam"}, False
    SRA_FASTQ = {}, False
    FASTA_QUAL = {}, False

    def __init__(self, exts: Set[str], can_output: bool):
        self.exts = exts
        self.can_output = can_output

    @staticmethod
    def guess_from_name(
        path: Union[str, Path, IO],
        output: bool,
        raise_on_failure: bool = False,
    ):
        """
            Detects file format based on the file name.

            Args:
                path: The filename to guess.
                output: Whether to only guess formats that can be output.
                raise_on_failure: Whether to raise an exception if the filename cannot
                    be detected.

            Returns:
                SequenceFileType.
            """
        name = None

        if isinstance(path, str):
            name = path
        elif hasattr(path, "name"):
            name = path.name

        if name:
            name, ext1, _ = splitext_compressed(name)
            ext = ext1.lower()

            for t in SequenceFileType:
                if output and not t.can_output:
                    continue
                if ext in t.exts:
                    return t

            if ext == ".txt" and name.endswith("_sequence"):
                return SequenceFileType.FASTQ

        if raise_on_failure:
            raise UnknownFileTypeError(
                f"Could not determine whether file {path!r} is FASTA, FASTQ, or SAM/BAM"
            )
